/* computer/apple.c++ -- written by Alexis WILKE (c) 2005-2006 */

/**********************************************************************
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**********************************************************************/


#include	"emu65xx.h"
#include	"screen.h"
#include	<fltk_window.h>
#include	<fltk/run.h>
#include	<fltk/rgbImage.h>
#include	<fltk/FileChooser.h>
#include	<fltk/ReturnButton.h>

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	<assert.h>
#include	<time.h>


emulator_8bits::emu65xx *		g_computer;
fltk::Window *				g_window;
fltk::Window *				g_about;
fltk::Window *				g_license;
fltk::Window *				g_copyright;
fltk::Window *				g_ask_address;
fltk::Window *				g_ask_speed;
fltk::Window *				g_ask_address_and_size;
fltk::rgbImage *			g_screen;
ScreenOutput *				g_screen_widget;
fltk::Item *				g_trace_item;
fltk::Item *				g_program_select_item;
fltk::Item *				g_text_select_item;
fltk::Item *				g_lores_select_item;
fltk::Item *				g_hires_select_item;
fltk::Item *				g_mixed_select_item;
fltk::Item *				g_screen_select_item;
fltk::Item *				g_statistics_item;
fltk::FileChooser *			g_file_chooser;
fltk::Output *				g_instructions_executed_item;
fltk::Output *				g_cycles_item;
fltk::Output *				g_memory_access_item;
fltk::Output *				g_reads_item;
fltk::Output *				g_writes_item;
fltk::Output *				g_page_access_item;
fltk::Output *				g_unknown_instructions_item;
fltk::Output *				g_instructions_used_item;
fltk::Output *				g_page_read_item;
fltk::Output *				g_page_write_item;
fltk::Item *				g_auto_reset_stats_item;
fltk::Item *				g_1mhz_item;
fltk::Item *				g_2mhz_item;
fltk::Item *				g_4mhz_item;
fltk::Item *				g_6mhz_item;
fltk::Item *				g_8mhz_item;
fltk::Item *				g_10mhz_item;
fltk::Item *				g_other_mhz_item;
fltk::Item *				g_30mhz_monitor_item;
fltk::Item *				g_20mhz_monitor_item;
fltk::Item *				g_10mhz_monitor_item;
//fltk::Output *				g_about_name;
//fltk::Output *				g_about_version;
//fltk::Output *				g_about_copyright;
//fltk::Output *				g_about_author;


int					g_zoom;
float					g_refresh_rate = 30.0f;




const char g_license_str[] =
	"		    GNU GENERAL PUBLIC LICENSE\n"
	"		       Version 2, June 1991\n"
	"\n"
	" Copyright (C) 1989, 1991 Free Software Foundation, Inc.\n"
	"			  675 Mass Ave, Cambridge, MA 02139, USA\n"
	" Everyone is permitted to copy and distribute verbatim copies "
	" of this license document, but changing it is not allowed.\n"
	"\n"
	"			    Preamble\n"
	"\n"
	"  The licenses for most software are designed to take away your "
	"freedom to share and change it.  By contrast, the GNU General Public "
	"License is intended to guarantee your freedom to share and change free "
	"software--to make sure the software is free for all its users.  This "
	"General Public License applies to most of the Free Software "
	"Foundation's software and to any other program whose authors commit to "
	"using it.  (Some other Free Software Foundation software is covered by "
	"the GNU Library General Public License instead.)  You can apply it to "
	"your programs, too.\n"
	"\n"
	"  When we speak of free software, we are referring to freedom, not "
	"price.  Our General Public Licenses are designed to make sure that you "
	"have the freedom to distribute copies of free software (and charge for "
	"this service if you wish), that you receive source code or can get it "
	"if you want it, that you can change the software or use pieces of it "
	"in new free programs; and that you know you can do these things.\n"
	"\n"
	"  To protect your rights, we need to make restrictions that forbid "
	"anyone to deny you these rights or to ask you to surrender the rights. "
	"These restrictions translate to certain responsibilities for you if you "
	"distribute copies of the software, or if you modify it.\n"
	"\n"
	"  For example, if you distribute copies of such a program, whether "
	"gratis or for a fee, you must give the recipients all the rights that "
	"you have.  You must make sure that they, too, receive or can get the "
	"source code.  And you must show them these terms so they know their "
	"rights.\n"
	"\n"
	"  We protect your rights with two steps: (1) copyright the software, and "
	"(2) offer you this license which gives you legal permission to copy, "
	"distribute and/or modify the software.\n"
	"\n"
	"  Also, for each author's protection and ours, we want to make certain "
	"that everyone understands that there is no warranty for this free "
	"software.  If the software is modified by someone else and passed on, we "
	"want its recipients to know that what they have is not the original, so "
	"that any problems introduced by others will not reflect on the original "
	"authors' reputations.\n"
	"\n"
	"  Finally, any free program is threatened constantly by software "
	"patents.  We wish to avoid the danger that redistributors of a free "
	"program will individually obtain patent licenses, in effect making the "
	"program proprietary.  To prevent this, we have made it clear that any "
	"patent must be licensed for everyone's free use or not licensed at all.\n"
	"\n"
	"  The precise terms and conditions for copying, distribution and "
	"modification follow.\n"
	"\n\n\n"
	"		    GNU GENERAL PUBLIC LICENSE\n"
	"   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION\n"
	"\n"
	"  0. This License applies to any program or other work which contains "
	"a notice placed by the copyright holder saying it may be distributed "
	"under the terms of this General Public License.  The \"Program\", below, "
	"refers to any such program or work, and a \"work based on the Program\" "
	"means either the Program or any derivative work under copyright law: "
	"that is to say, a work containing the Program or a portion of it, "
	"either verbatim or with modifications and/or translated into another "
	"language.  (Hereinafter, translation is included without limitation in "
	"the term \"modification\".)  Each licensee is addressed as \"you\".\n"
	"\n"
	"Activities other than copying, distribution and modification are not "
	"covered by this License; they are outside its scope.  The act of "
	"running the Program is not restricted, and the output from the Program "
	"is covered only if its contents constitute a work based on the "
	"Program (independent of having been made by running the Program). "
	"Whether that is true depends on what the Program does.\n"
	"\n"
	"  1. You may copy and distribute verbatim copies of the Program's "
	"source code as you receive it, in any medium, provided that you "
	"conspicuously and appropriately publish on each copy an appropriate "
	"copyright notice and disclaimer of warranty; keep intact all the "
	"notices that refer to this License and to the absence of any warranty; "
	"and give any other recipients of the Program a copy of this License "
	"along with the Program.\n"
	"\n"
	"You may charge a fee for the physical act of transferring a copy, and "
	"you may at your option offer warranty protection in exchange for a fee.\n"
	"\n"
	"  2. You may modify your copy or copies of the Program or any portion "
	"of it, thus forming a work based on the Program, and copy and "
	"distribute such modifications or work under the terms of Section 1 "
	"above, provided that you also meet all of these conditions:\n"
	"\n"
	"    a) You must cause the modified files to carry prominent notices "
	"stating that you changed the files and the date of any change.\n"
	"\n"
	"    b) You must cause any work that you distribute or publish, that in "
	"whole or in part contains or is derived from the Program or any "
	"part thereof, to be licensed as a whole at no charge to all third "
	"parties under the terms of this License.\n"
	"\n"
	"    c) If the modified program normally reads commands interactively "
	"when run, you must cause it, when started running for such "
	"interactive use in the most ordinary way, to print or display an "
	"announcement including an appropriate copyright notice and a "
	"notice that there is no warranty (or else, saying that you provide "
	"a warranty) and that users may redistribute the program under "
	"these conditions, and telling the user how to view a copy of this "
	"License.  (Exception: if the Program itself is interactive but "
	"does not normally print such an announcement, your work based on "
	"the Program is not required to print an announcement.)\n"
	"\n\n\n"
	"These requirements apply to the modified work as a whole.  If "
	"identifiable sections of that work are not derived from the Program, "
	"and can be reasonably considered independent and separate works in "
	"themselves, then this License, and its terms, do not apply to those "
	"sections when you distribute them as separate works.  But when you "
	"distribute the same sections as part of a whole which is a work based "
	"on the Program, the distribution of the whole must be on the terms of "
	"this License, whose permissions for other licensees extend to the "
	"entire whole, and thus to each and every part regardless of who wrote it.\n"
	"\n"
	"Thus, it is not the intent of this section to claim rights or contest "
	"your rights to work written entirely by you; rather, the intent is to "
	"exercise the right to control the distribution of derivative or "
	"collective works based on the Program.\n"
	"\n"
	"In addition, mere aggregation of another work not based on the Program "
	"with the Program (or with a work based on the Program) on a volume of "
	"a storage or distribution medium does not bring the other work under "
	"the scope of this License.\n"
	"\n"
	"  3. You may copy and distribute the Program (or a work based on it, "
	"under Section 2) in object code or executable form under the terms of "
	"Sections 1 and 2 above provided that you also do one of the following:\n"
	"\n"
	"    a) Accompany it with the complete corresponding machine-readable "
	"source code, which must be distributed under the terms of Sections "
	"1 and 2 above on a medium customarily used for software interchange; or,\n"
	"\n"
	"    b) Accompany it with a written offer, valid for at least three "
	"years, to give any third party, for a charge no more than your "
	"cost of physically performing source distribution, a complete "
	"machine-readable copy of the corresponding source code, to be "
	"distributed under the terms of Sections 1 and 2 above on a medium "
	"customarily used for software interchange; or,\n"
	"\n"
	"    c) Accompany it with the information you received as to the offer "
	"to distribute corresponding source code.  (This alternative is "
	"allowed only for noncommercial distribution and only if you "
	"received the program in object code or executable form with such "
	"an offer, in accord with Subsection b above.)\n"
	"\n"
	"The source code for a work means the preferred form of the work for "
	"making modifications to it.  For an executable work, complete source "
	"code means all the source code for all modules it contains, plus any "
	"associated interface definition files, plus the scripts used to "
	"control compilation and installation of the executable.  However, as a "
	"special exception, the source code distributed need not include "
	"anything that is normally distributed (in either source or binary "
	"form) with the major components (compiler, kernel, and so on) of the "
	"operating system on which the executable runs, unless that component "
	"itself accompanies the executable.\n"
	"\n"
	"If distribution of executable or object code is made by offering "
	"access to copy from a designated place, then offering equivalent "
	"access to copy the source code from the same place counts as "
	"distribution of the source code, even though third parties are not "
	"compelled to copy the source along with the object code.\n"
	"\n\n\n"
	"  4. You may not copy, modify, sublicense, or distribute the Program "
	"except as expressly provided under this License.  Any attempt "
	"otherwise to copy, modify, sublicense or distribute the Program is "
	"void, and will automatically terminate your rights under this License. "
	"However, parties who have received copies, or rights, from you under "
	"this License will not have their licenses terminated so long as such "
	"parties remain in full compliance.\n"
	"\n"
	"  5. You are not required to accept this License, since you have not "
	"signed it.  However, nothing else grants you permission to modify or "
	"distribute the Program or its derivative works.  These actions are "
	"prohibited by law if you do not accept this License.  Therefore, by "
	"modifying or distributing the Program (or any work based on the "
	"Program), you indicate your acceptance of this License to do so, and "
	"all its terms and conditions for copying, distributing or modifying "
	"the Program or works based on it.\n"
	"\n"
	"  6. Each time you redistribute the Program (or any work based on the "
	"Program), the recipient automatically receives a license from the "
	"original licensor to copy, distribute or modify the Program subject to "
	"these terms and conditions.  You may not impose any further "
	"restrictions on the recipients' exercise of the rights granted herein. "
	"You are not responsible for enforcing compliance by third parties to "
	"this License.\n"
	"\n"
	"  7. If, as a consequence of a court judgment or allegation of patent "
	"infringement or for any other reason (not limited to patent issues), "
	"conditions are imposed on you (whether by court order, agreement or "
	"otherwise) that contradict the conditions of this License, they do not "
	"excuse you from the conditions of this License.  If you cannot "
	"distribute so as to satisfy simultaneously your obligations under this "
	"License and any other pertinent obligations, then as a consequence you "
	"may not distribute the Program at all.  For example, if a patent "
	"license would not permit royalty-free redistribution of the Program by "
	"all those who receive copies directly or indirectly through you, then "
	"the only way you could satisfy both it and this License would be to "
	"refrain entirely from distribution of the Program.\n"
	"\n"
	"If any portion of this section is held invalid or unenforceable under "
	"any particular circumstance, the balance of the section is intended to "
	"apply and the section as a whole is intended to apply in other "
	"circumstances.\n"
	"\n"
	"It is not the purpose of this section to induce you to infringe any "
	"patents or other property right claims or to contest validity of any "
	"such claims; this section has the sole purpose of protecting the "
	"integrity of the free software distribution system, which is "
	"implemented by public license practices.  Many people have made "
	"generous contributions to the wide range of software distributed "
	"through that system in reliance on consistent application of that "
	"system; it is up to the author/donor to decide if he or she is willing "
	"to distribute software through any other system and a licensee cannot "
	"impose that choice.\n"
	"\n"
	"This section is intended to make thoroughly clear what is believed to "
	"be a consequence of the rest of this License.\n"
	"\n\n\n"
	"  8. If the distribution and/or use of the Program is restricted in "
	"certain countries either by patents or by copyrighted interfaces, the "
	"original copyright holder who places the Program under this License "
	"may add an explicit geographical distribution limitation excluding "
	"those countries, so that distribution is permitted only in or among "
	"countries not thus excluded.  In such case, this License incorporates "
	"the limitation as if written in the body of this License.\n"
	"\n"
	"  9. The Free Software Foundation may publish revised and/or new versions "
	"of the General Public License from time to time.  Such new versions will "
	"be similar in spirit to the present version, but may differ in detail to "
	"address new problems or concerns.\n"
	"\n"
	"Each version is given a distinguishing version number.  If the Program "
	"specifies a version number of this License which applies to it and \"any "
	"later version\", you have the option of following the terms and conditions "
	"either of that version or of any later version published by the Free "
	"Software Foundation.  If the Program does not specify a version number of "
	"this License, you may choose any version ever published by the Free Software "
	"Foundation.\n"
	"\n"
	"  10. If you wish to incorporate parts of the Program into other free "
	"programs whose distribution conditions are different, write to the author "
	"to ask for permission.  For software which is copyrighted by the Free "
	"Software Foundation, write to the Free Software Foundation; we sometimes "
	"make exceptions for this.  Our decision will be guided by the two goals "
	"of preserving the free status of all derivatives of our free software and"
	"of promoting the sharing and reuse of software generally.\n"
	"\n"
	"			    NO WARRANTY\n"
	"\n"
	"  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY"
	"FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN"
	"OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES"
	"PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED"
	"OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF"
	"MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS"
	"TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE"
	"PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, "
	"REPAIR OR CORRECTION.\n"
	"\n"
	"  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING "
	"WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR "
	"REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, "
	"INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING "
	"OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED "
	"TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY "
	"YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER "
	"PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE "
	"POSSIBILITY OF SUCH DAMAGES.\n"
	"\n"
	"		     END OF TERMS AND CONDITIONS\n"
	"\n\n\n"
	"	Appendix: How to Apply These Terms to Your New Programs\n"
	"\n"
	"  If you develop a new program, and you want it to be of the greatest "
	"possible use to the public, the best way to achieve this is to make it "
	"free software which everyone can redistribute and change under these terms.\n"
	"\n"
	"  To do so, attach the following notices to the program.  It is safest "
	"to attach them to the start of each source file to most effectively "
	"convey the exclusion of warranty; and each file should have at least "
	"the \"copyright\" line and a pointer to where the full notice is found.\n"
	"\n"
	"    <one line to give the program's name and a brief idea of what it does.>\n"
	"    Copyright (C) 19yy  <name of author>\n"
	"\n"
	"    This program is free software; you can redistribute it and/or modify "
	"it under the terms of the GNU General Public License as published by "
	"the Free Software Foundation; either version 2 of the License, or "
	"(at your option) any later version.\n"
	"\n"
	"    This program is distributed in the hope that it will be useful, "
	"but WITHOUT ANY WARRANTY; without even the implied warranty of "
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the "
	"GNU General Public License for more details.\n"
	"\n"
	"    You should have received a copy of the GNU General Public License "
	"along with this program; if not, write to the Free Software "
	"Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n"
	"\n"
	"Also add information on how to contact you by electronic and paper mail.\n"
	"\n"
	"If the program is interactive, make it output a short notice like this "
	"when it starts in an interactive mode:\n"
	"\n"
	"    Gnomovision version 69, Copyright (C) 19yy name of author\n"
	"    Gnomovision comes with ABSOLUTELY NO WARRANTY; for details type `show w'.\n"
	"    This is free software, and you are welcome to redistribute it\n"
	"    under certain conditions; type `show c' for details.\n"
	"\n"
	"The hypothetical commands `show w' and `show c' should show the appropriate "
	"parts of the General Public License.  Of course, the commands you use may "
	"be called something other than `show w' and `show c'; they could even be "
	"mouse-clicks or menu items--whatever suits your program.\n"
	"\n"
	"You should also get your employer (if you work as a programmer) or your "
	"school, if any, to sign a \"copyright disclaimer\" for the program, if "
	"necessary.  Here is a sample; alter the names:\n"
	"\n"
	"  Yoyodyne, Inc., hereby disclaims all copyright interest in the program "
	"`Gnomovision' (which makes passes at compilers) written by James Hacker.\n"
	"\n"
	"  <signature of Ty Coon>, 1 April 1989\n"
	"  Ty Coon, President of Vice\n"
	"\n"
	"This General Public License does not permit incorporating your program into "
	"proprietary programs.  If your program is a subroutine library, you may "
	"consider it more useful to permit linking proprietary applications with the "
	"library.  If this is what you want to do, use the GNU Library General "
	"Public License instead of this License.\n"
;



const char g_copyright_str[] =
	"The Apple Emulator software is a copyright by its author Alexis Wilke.\n"
	"\n"
	"The software was created in 2005-2006. Some very minor parts are based "
	"on work done by others, however, the entire system is a complete redesigned "
	"by Alexis Wilke.\n"
	"\n"
	"This software uses the FLTK library to render its output.\n"
	"\n"
	"Copyright (c) 2005-2006  Alexis Wilke.\n"
;



class Apple : public emulator_8bits::emu65xx::Callback
{
public:
	static const unsigned int	VIDEO_MODE_PAGE0     = 0x00;	// page 0 (400-7FF or 2000-3FFF)
	static const unsigned int	VIDEO_MODE_PAGE1     = 0x01;	// page 1 (800-BFF or 4000-5FFF)
	static const unsigned int	VIDEO_MODE_PAGE_MASK = 0x01;

	static const unsigned int	VIDEO_MODE_TEXT     = 0x00;	// all text
	static const unsigned int	VIDEO_MODE_GRAPHICS = 0x02;	// all graphics
	static const unsigned int	VIDEO_MODE_MIXED    = 0x04;	// 4 lines of text at the bottom
	static const unsigned int	VIDEO_MODE_MASK     = 0x06;

	static const unsigned int	VIDEO_MODE_LORES    = 0x00;	// using text info
	static const unsigned int	VIDEO_MODE_HIRES    = 0x08;	// using hires memory
	static const unsigned int	VIDEO_MODE_RES_MASK = 0x08;

	static const unsigned int	VIDEO_MODE_SINGLE     = 0x00;	// 40 columns
	static const unsigned int	VIDEO_MODE_DOUBLE     = 0x10;	// 80 columns
	static const unsigned int	VIDEO_MODE_WIDTH_MASK = 0x10;

	static const unsigned int	VIDEO_MODE_PROGRAM_SELECT = 0x00;	// standard Apple ][ software selection
	static const unsigned int	VIDEO_MODE_USER_SELECT    = 0x20;	// the user set up the video mode via the menu
	static const unsigned int	VIDEO_MODE_SELECT_MASK    = 0x20;

	struct rgb_t {
		unsigned char	red;
		unsigned char	green;
		unsigned char	blue;
	};

				Apple(void);
	virtual			~Apple();

	void			LoadROM(const char *name, unsigned short addr, unsigned short size);
	void			LoadCharacters(const char *name);
	void			LoadFile(const char *filename, long ptr, bool basic = false);
	void			SaveFile(const char *filename, long ptr, long len, bool basic = false);
	void			PatchROM(unsigned short addr, unsigned char code);

	virtual void		OnNMI(void);
	virtual void		OnReset(void);
	virtual void		OnInterrupt(void);

	virtual void		OnLoad(unsigned short addr, unsigned char& value);
	virtual void		OnStore(unsigned short addr, unsigned char value);

	virtual void		OnInstruction(void);
	virtual void		OnUnknownOpcode(unsigned short& pc);
	virtual void		OnTrace(const char *msg);

	void			SetKey(unsigned char key);
	bool			CreateRGBImage(rgb_t *rgb, int zoom, bool all);
	void			SelectUserVideo(unsigned int mode);

private:
	static const unsigned int	CHANGED_TEXT       = 0x01;
	static const unsigned int	CHANGED_FULL_TEXT  = 0x02;
	static const unsigned int	CHANGED_LORES      = 0x04;
	static const unsigned int	CHANGED_FULL_LORES = 0x08;
	static const unsigned int	CHANGED_HIRES      = 0x10;
	static const unsigned int	CHANGED_FULL_HIRES = 0x20;
	static const unsigned int	CHANGED_ALL        = 0x3F;

	void			ResetMap(void);
	void			DrawChar(unsigned short addr, unsigned char c);
	void			DrawPixels(unsigned short addr, unsigned char c0, unsigned char c1);
	void			ChangedVideoXY(unsigned int *array, int x, int y);
	void			GetVideoXY(unsigned short addr, int& x, int& y);
	bool			DrawText(rgb_t *dst, const unsigned int *src, unsigned int repeat, unsigned int *changed, int zoom, bool all);
	bool			DrawLoRes(rgb_t *dst, const unsigned int *src, unsigned int repeat, unsigned int *changed, int zoom, bool all);
	bool			DrawHiRes(rgb_t *dst, const unsigned int *src, unsigned int repeat, unsigned int *changed, int zoom, bool all);
	void			SelectVideo(int mode);

	unsigned char		f_memory[0x60000];	// 6 x 64Kb (Apple //c max.)
	unsigned char *		f_map[16];
	unsigned int		f_text_colors[2];	// black & white by default, you can use green, etc. if you wish
	unsigned int		f_lores_colors[16];	// see below
	unsigned int		f_hires_colors[8];	// see below (not really 8 colors!)
	unsigned int		f_changed;
	unsigned int		f_text_changed[40 * 24 / 32 * 2];	// the two whole screens of text
	unsigned int		f_lores_changed[40 * 48 / 32 * 2];	// the two whole screens of lores
	unsigned int		f_hires_changed[40 * 192 / 32 * 2];	// the two whole screens of hires
	unsigned int		f_lores_bitmaps[40 * 48 * 2];
	unsigned int		f_text_bitmaps[40 * 7 * 24 * 8 * 2];
	unsigned int		f_hires_bitmaps[40 * 7 * 24 * 8 * 2];
	unsigned char		f_characters[256 * 8];
	bool			f_black_n_white;
	unsigned int		f_video_mode;		// see VIDEO_MODE_...
	unsigned int		f_user_video_mode;	// menu selected video mode
};

Apple::rgb_t *					g_screen_data;



Apple::Apple(void)
{
	int		i;

	// randomize the memory at the start -- usually, on this hardware
	// it was very random
#if 1
	srand(time(NULL));
	for(i = 0; i < sizeof(f_memory); i++) {
		f_memory[i] = rand();
	}
	// C000 to C010 inclusive is cleared (i.e. keyboard data)
	// [this is done by a proper ROM]
	//for(i = 0; i <= 16; i++) {
	//	f_memory[0xC000 + i] &= 0x7F;
	//}
#else
	// the Apple ][ ROM I have seems to expect a reset memory?!
	memset(f_memory, 0, sizeof(f_memory));
#endif

	ResetMap();

// define the standard colors
	f_text_colors[0] = 0x000000;	// Black (could be not 100% black..)
	f_text_colors[1] = 0xFFFFFF;	// White (could be green!)

	f_lores_colors[ 0] = 0x000000;	// Black
	f_lores_colors[ 1] = 0xA54242;	// Magenta
	f_lores_colors[ 2] = 0x0000A0;	// Dark Blue
	f_lores_colors[ 3] = 0x8A008A;	// Purple
	f_lores_colors[ 4] = 0x006400;	// Dark Green
	f_lores_colors[ 5] = 0x808080;	// Grey
	f_lores_colors[ 6] = 0x1C86FF;	// Medium Blue
	f_lores_colors[ 7] = 0x41FFFF;	// Light Blue
	f_lores_colors[ 8] = 0x8A8A00;	// Brown
	f_lores_colors[ 9] = 0xFF4200;	// Orange
	f_lores_colors[10] = 0x5A5A5A;	// Dark Grey
	f_lores_colors[11] = 0xFFC6CB;	// Pink
	f_lores_colors[12] = 0x00FF00;	// Green
	f_lores_colors[13] = 0xFFFF00;	// Yellow
	f_lores_colors[14] = 0xA0FFA0;	// Aqua
	f_lores_colors[15] = 0xFFFFFF;	// White

	f_hires_colors[0] = 0x000000;	// Black
	f_hires_colors[1] = 0x00FF00;	// Green
	f_hires_colors[2] = 0x8A008A;	// Purple
	f_hires_colors[3] = 0xFFFFFF;	// White
	f_hires_colors[4] = 0x000000;	// Black
	f_hires_colors[5] = 0xFF0000;	// Red
	f_hires_colors[6] = 0x0000FF;	// Blue
	f_hires_colors[7] = 0xFFFFFF;	// White

	f_changed = CHANGED_ALL;
	memset(f_text_changed,  -1, sizeof(f_text_changed));
	memset(f_lores_changed, -1, sizeof(f_lores_changed));
	memset(f_hires_changed, -1, sizeof(f_hires_changed));

	memset(f_lores_bitmaps, -1, sizeof(f_lores_bitmaps));
	memset(f_text_bitmaps,  -1, sizeof(f_text_bitmaps));
	memset(f_hires_bitmaps, -1, sizeof(f_hires_bitmaps));

	f_black_n_white = false;
	f_video_mode = 0;
	f_user_video_mode = 0;
}


Apple::~Apple()
{
	delete [] f_memory;
}


void Apple::LoadROM(const char *name, unsigned short addr, unsigned short size)
{
	FILE *f = fopen(name, "rb");
	if(f == 0) {
		fprintf(stderr, "ERROR: can't load file \"%s\".\n", name);
		exit(1);
	}
	if(fread(f_memory + addr, 1, size, f) != size) {
		fprintf(stderr, "ERROR: can't read %d bytes from file \"%s\".\n", size, name);
		exit(1);
	}

	printf("Loaded ROM \"%s\"\n", name);
}


void Apple::PatchROM(unsigned short addr, unsigned char code)
{
	f_memory[addr] = code;
}


void Apple::LoadCharacters(const char *name)
{
	FILE *f = fopen(name, "rb");
	if(f == 0) {
		fprintf(stderr, "ERROR: can't load file \"%s\".\n", name);
		exit(1);
	}
	if(fread(f_characters, 1, 2048, f) != 2048) {
		fprintf(stderr, "ERROR: can't read %d bytes from file \"%s\".\n", 2048, name);
		exit(1);
	}

int i;
for(i = 0; i < 64 * 8; i++) f_characters[i] ^= 255;
// the following are flashing characters, we need to do something about that...
for(; i < 64 * 8 * 2; i++) if(i & 4) f_characters[i] ^= 255;

	printf("Loaded Characters \"%s\"\n", name);
}


void Apple::LoadFile(const char *filename, long ptr, bool basic)
{
	FILE *f = fopen(filename, "rb");
	if(f == 0) {
		// TODO: display an error message
		return;
	}
	fseek(f, 0, SEEK_END);
	long size = ftell(f);
	fseek(f, 0, SEEK_SET);
	if(ptr < 0 || ptr + size > sizeof(f_memory)) {
		fclose(f);
		// TODO: display an error message
		return;
	}
	if(basic) {
		f_memory[0x800] = 0;	// not too sure about that one...
		char buf[8];
		fread(buf, 1, 8, f);
		if(memcmp(buf, "APPBASIC", 8) != 0) {
			// TODO: display an error message
			fprintf(stderr, "Error: APPBASIC not found at the start of the file\n");
			return;
		}
		fread(f_memory + 0x67, 1, 8, f);
		fread(f_memory + 0xAF, 1, 2, f);
		size -= 18;
		ptr = f_memory[0x67] + f_memory[0x68] * 256;
		if(ptr < 0 || ptr + size > sizeof(f_memory)) {
			fclose(f);
			// TODO: display an error message
			fprintf(stderr, "Error: pointer out of bounds\n");
			return;
		}
	}
	// load it straight to the memory!
	fread(f_memory + ptr, 1, size, f);
	fclose(f);
}


void Apple::SaveFile(const char *filename, long ptr, long size, bool basic)
{
	// we save the basic program as defined including the final 0x00 0x00
	if(basic) {
		ptr  = f_memory[0x67] + f_memory[0x68] * 256;
		size = f_memory[0xAF] + f_memory[0xB0] * 256 - ptr;
	}

	if(ptr > sizeof(f_memory)
	|| ptr + size > sizeof(f_memory)) {
		// TODO: display an error message
		return;
	}
	FILE *f = fopen(filename, "wb");
	if(f == 0) {
		// TODO: display an error message
		return;
	}
	if(basic) {
		// save the location so on reload we do it at the right place
		fwrite("APPBASIC", 1, 8, f);		// a magic #
		fwrite(f_memory + 0x67, 1, 8, f);
		fwrite(f_memory + 0xAF, 1, 2, f);
	}
	// load it straight to the memory!
	fwrite(f_memory + ptr, 1, size, f);
	fclose(f);
}


void Apple::OnNMI(void)
{
	ResetMap();
}

void Apple::OnReset(void)
{
	ResetMap();
}

void Apple::OnInterrupt(void)
{
	ResetMap();
}


void Apple::SelectVideo(int mode)
{
	unsigned int		size, *changed;

	size = 0;

	// TODO: this needs to be thread safe
	switch(mode) {
	case 0:
		if((f_video_mode & VIDEO_MODE_GRAPHICS) == 0) {
			f_video_mode |= VIDEO_MODE_GRAPHICS;
			if((f_video_mode & VIDEO_MODE_HIRES) != 0) {
				if((f_video_mode & VIDEO_MODE_PAGE1) != 0) {
					changed = f_hires_changed + 40 * 192 / 32;
				}
				else {
					changed = f_hires_changed;
				}
				if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
					size = 40 * 192 / 32;
					f_changed |= CHANGED_HIRES | CHANGED_FULL_HIRES;
				}
				else {
					size = 40 * 160 / 32;
					f_changed |= CHANGED_HIRES;
				}
			}
			else {
				if((f_video_mode & VIDEO_MODE_PAGE1) != 0) {
					changed = f_lores_changed + 40 * 48 / 32;
				}
				else {
					changed = f_lores_changed;
				}
				if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
					size = 40 * 48 / 32;
					f_changed |= CHANGED_LORES | CHANGED_FULL_LORES;
				}
				else {
					size = 40 * 40 / 32;
					f_changed |= CHANGED_LORES;
				}
			}
		}
		break;

	case 1:
		if((f_video_mode & VIDEO_MODE_GRAPHICS) != 0) {
			f_video_mode &= ~VIDEO_MODE_GRAPHICS;
			if((f_video_mode & VIDEO_MODE_PAGE1) != 0) {
				changed = f_text_changed + 40 * 24 / 32;
			}
			else {
				changed = f_text_changed;
			}
			if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
				f_changed |= CHANGED_TEXT | CHANGED_FULL_TEXT;
				size = 40 * 24 / 32;
			}
			else {
				f_changed |= CHANGED_TEXT;
				size = 40 * 20 / 32;
			}
		}
		break;

	case 2:
		if((f_video_mode & VIDEO_MODE_MIXED) != 0) {
			f_video_mode &= ~VIDEO_MODE_MIXED;
			if((f_video_mode & VIDEO_MODE_GRAPHICS) != 0) {
				f_video_mode |= VIDEO_MODE_GRAPHICS;
				if((f_video_mode & VIDEO_MODE_HIRES) != 0) {
					if((f_video_mode & VIDEO_MODE_PAGE1) != 0) {
						changed = f_hires_changed + 40 * 160 / 32 + 40 * 192 / 32;
					}
					else {
						changed = f_hires_changed + 40 * 160 / 32;
					}
					size = 40 * 32 / 32;
					f_changed |= CHANGED_FULL_HIRES;
				}
				else {
					if((f_video_mode & VIDEO_MODE_PAGE1) != 0) {
						changed = f_lores_changed + 40 * 40 / 32 + 40 * 48 / 32;
					}
					else {
						changed = f_lores_changed + 40 * 40 / 32;
					}
					size = 40 * 8 / 32;
					f_changed |= CHANGED_FULL_LORES;
				}
			}
		}
		break;

	case 3:
		if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
			f_video_mode |= VIDEO_MODE_MIXED;
			if((f_video_mode & VIDEO_MODE_PAGE1) != 0) {
				changed = f_text_changed + 40 * 24 / 32 + 40 * 20 / 32;
			}
			else {
				changed = f_text_changed + 40 * 20 / 32;
			}
			f_changed |= CHANGED_FULL_TEXT;
			size = 40 * 4 / 32;
		}
		break;

	case 4:
		if((f_video_mode & VIDEO_MODE_PAGE1) != 0) {
			f_video_mode &= ~VIDEO_MODE_PAGE1;
			if((f_video_mode & VIDEO_MODE_GRAPHICS) != 0) {
				if((f_video_mode & VIDEO_MODE_HIRES) != 0) {
					changed = f_hires_changed;
					if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
						size = 40 * 192 / 32;
						f_changed |= CHANGED_HIRES | CHANGED_FULL_HIRES;
					}
					else {
						size = 40 * 160 / 32;
						f_changed |= CHANGED_HIRES;
					}
				}
				else {
					changed = f_lores_changed;
					if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
						size = 40 * 48 / 32;
						f_changed |= CHANGED_LORES | CHANGED_FULL_LORES;
					}
					else {
						size = 40 * 40 / 32;
						f_changed |= CHANGED_LORES;
					}
				}
				if((f_video_mode & VIDEO_MODE_MIXED) != 0) {
					memset(f_text_changed + 40 * 20 / 32, -1, (40 * 4 / 32) * sizeof(unsigned int));
					f_changed |= CHANGED_FULL_TEXT;
				}
			}
			else {
				changed = f_text_changed;
				size = 40 * 24 / 32;
				f_changed |= CHANGED_TEXT | CHANGED_FULL_TEXT;
			}
		}
		break;

	case 5:
		if((f_video_mode & VIDEO_MODE_PAGE1) == 0) {
			f_video_mode |= VIDEO_MODE_PAGE1;
			if((f_video_mode & VIDEO_MODE_GRAPHICS) != 0) {
				if((f_video_mode & VIDEO_MODE_HIRES) != 0) {
					changed = f_hires_changed + 40 * 192 / 32;
					if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
						size = 40 * 192 / 32;
						f_changed |= CHANGED_HIRES | CHANGED_FULL_HIRES;
					}
					else {
						size = 40 * 160 / 32;
						f_changed |= CHANGED_HIRES;
					}
				}
				else {
					changed = f_lores_changed + 40 * 48 / 32;
					if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
						size = 40 * 48 / 32;
						f_changed |= CHANGED_LORES | CHANGED_FULL_LORES;
					}
					else {
						size = 40 * 40 / 32;
						f_changed |= CHANGED_LORES;
					}
				}
				if((f_video_mode & VIDEO_MODE_MIXED) != 0) {
					memset(f_text_changed + 40 * 20 / 32 + 40 * 24 / 32, -1, (40 * 4 / 32) * sizeof(unsigned int));
					f_changed |= CHANGED_FULL_TEXT;
				}
			}
			else {
				changed = f_text_changed + 40 * 24 / 32;
				size = 40 * 24 / 32;
				f_changed |= CHANGED_TEXT | CHANGED_FULL_TEXT;
			}
		}
		break;

	case 6:
		if((f_video_mode & VIDEO_MODE_HIRES) != 0) {
			f_video_mode &= ~VIDEO_MODE_HIRES;
			if((f_video_mode & VIDEO_MODE_GRAPHICS) != 0) {
				if((f_video_mode & VIDEO_MODE_PAGE1) != 0) {
					changed = f_lores_changed + 40 * 48 / 32;
				}
				else {
					changed = f_lores_changed;
				}
				if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
					size = 40 * 48 / 32;
					f_changed |= CHANGED_LORES | CHANGED_FULL_LORES;
				}
				else {
					size = 40 * 40 / 32;
					f_changed |= CHANGED_LORES;
				}
			}
		}
		break;

	case 7:
		if((f_video_mode & VIDEO_MODE_HIRES) == 0) {
			f_video_mode |= VIDEO_MODE_HIRES;
			if((f_video_mode & VIDEO_MODE_GRAPHICS) != 0) {
				if((f_video_mode & VIDEO_MODE_PAGE1) != 0) {
					changed = f_hires_changed + 40 * 192 / 32;
				}
				else {
					changed = f_hires_changed;
				}
				if((f_video_mode & VIDEO_MODE_MIXED) == 0) {
					size = 40 * 192 / 32;
					f_changed |= CHANGED_HIRES | CHANGED_FULL_HIRES;
				}
				else {
					size = 40 * 160 / 32;
					f_changed |= CHANGED_HIRES;
				}
			}
		}
		break;

	}

	if(size != 0) {
		memset(changed, -1, size * sizeof(unsigned int));
	}
}


void Apple::SelectUserVideo(unsigned int mode)
{
	f_user_video_mode = mode;

	// this could be as complicated as the previous function, however,
	// since it will be called only when the user selects one of the view
	// options, it won't hurt to force a full redraw
	f_changed = CHANGED_ALL;
	memset(f_text_changed,  -1, sizeof(f_text_changed));
	memset(f_lores_changed, -1, sizeof(f_lores_changed));
	memset(f_hires_changed, -1, sizeof(f_hires_changed));

	// we want to make sure the entire screen is refreshed
	g_screen_widget->set_damage(1);
}


void Apple::OnLoad(unsigned short addr, unsigned char& value)
{
	unsigned char	c, *ptr;
	unsigned short	m;
	int		i;

	// Defines the map (every 4Kb)
	m = addr >> 12;

	ptr = f_map[m] + (addr & ((1 << 12) - 1));

	// NOTE: reading the text & graphics pages has no special special effect
	if(m == 0x0C) {		// I/O
		if(addr < 0xC100) {
			// That's the I/O on the Apple II
			switch(addr) {
			case 0xC000 ... 0xC00F:
				value = *ptr;
				// slow down the processor!
				// (we need to use the stats and wait corresponding
				// to the cycles spent)
				//usleep(1);
				return;

			case 0xC010:	// Keyboard
				// this is nearly like normal memory, it clears
				// bit 7 after being read
				value = *ptr;
				for(i = 0; i <= 16; ++i) {
					f_memory[0xC000 + i] &= 0x7F;
				}
				return;

			case 0xC050 ... 0xC057:
				SelectVideo(addr & 7);
				break;

			}
		}
		// other 0xC100 to 0xCFFF is expension slots ROM; read as is...
	}

	value = *ptr;
}

void Apple::OnStore(unsigned short addr, unsigned char value)
{
	unsigned char	c, *ptr;
	unsigned short	m;

	// Defines the map (every 4Kb)
	m = addr >> 12;

	/*
	if(m >= 0x0D) {
		// TODO: I think you could copy the ROM in RAM
		// with a simple loop as follow. This means we
		// need to pretend we are seeing the RAM and
		// not just skip this Store(). This is called
		// shadow RAM if I'm correct.
		//
		//	LDA #$D0
		//	STA $11
		//	LDY #$00
		//	STY $10
		//	LDX #$30
		// copy:
		//	LDA ($10),Y
		//	STA ($10),y
		//	INY
		//	BNE copy
		//	INC $11
		//	DEX
		//	BNE copy
		return;
	}
	*/

	ptr = f_map[m] + (addr & ((1 << 12) - 1));

	if(m == 0x0C) {
		if(addr < 0xC100) {
			// That's the I/O for the Apple II
			switch(addr) {
			case 0xC010:	// Keyboard
				// ignored
				return;

			}
		}
		// other 0xC100 to 0xCFFF is expension slots ROM
		// we ignore writes to it
		return;
	}

	// TODO: wouldn't this let us write in ROM?!
	// if shadow RAM exists, we should have two maps: a read and a write map
	*ptr = value;

	switch(m) {
	case 0x00:		// text screen is around there
		// WARNING: this depends on the current memory
		// being selected; under an Apple //c you had
		// the following modes:
		//	page 0, 1, 2 and 3 in 40 columns
		//	page 0, 1, 2 and 3 in 80 columns
		if(addr >= 0x400 && addr < 0xBF8
		&& (addr & 0x07F) < 0x78) {
			// text screen!
			DrawChar(addr, value);
		}
		break;

	case 0x02:		// graphics page 0/0
	case 0x03:		// graphics page 0/1
	case 0x04:		// graphics page 1/0
	case 0x05:		// graphics page 1/1
		if((addr & 0x07F) < 0x78) {
			ptr = (unsigned char *) ((long) ptr & -2);
			DrawPixels(addr, ptr[0], ptr[1]);
		}
		break;

	}
}


void Apple::OnInstruction(void)
{
	static unsigned int	last_cycles;

	emulator_8bits::emu65xx::cpu_stats_t stats;
	g_computer->GetStats(stats, false);

	/* limit the refresh */
	if(stats.f_cycles - last_cycles < 1000000) {
		return;
	}

	if(g_auto_reset_stats_item->value() != 0) {
		emulator_8bits::emu65xx::cpu_stats_t ignore;
		g_computer->GetStats(ignore);
		last_cycles = 0;
	}
	else {
		last_cycles = stats.f_cycles;
	}

	if(g_statistics_item->value()) {
		static char str[32];

		sprintf(str, "%'d", stats.f_instructions);
		g_instructions_executed_item->value(str);

		sprintf(str, "%d", stats.f_inst_used_count);
		g_instructions_used_item->value(str);

		sprintf(str, "%d", stats.f_unknown_opcodes);
		g_unknown_instructions_item->value(str);

		sprintf(str, "%'d", stats.f_read + stats.f_write);
		g_memory_access_item->value(str);

		sprintf(str, "%d", stats.f_page_read_count + stats.f_page_written_count);
		g_page_access_item->value(str);

		sprintf(str, "%'d", stats.f_cycles);
		g_cycles_item->value(str);

		sprintf(str, "%'d", stats.f_read);
		g_reads_item->value(str);

		sprintf(str, "%'d", stats.f_write);
		g_writes_item->value(str);

		sprintf(str, "%d", stats.f_page_read_count);
		g_page_read_item->value(str);

		sprintf(str, "%d", stats.f_page_written_count);
		g_page_write_item->value(str);
	}
}

void ResetStats(fltk::Item *, void *)
{
	emulator_8bits::emu65xx::cpu_stats_t ignore;
	g_computer->GetStats(ignore);
}

void Apple::OnUnknownOpcode(unsigned short& pc)
{
}

void Apple::OnTrace(const char *msg)
{
	printf("%s\n", msg);
}


void SetMhz(fltk::Item*, long mhz)
{
	if(mhz == 0) {
		char buf[16];
		snprintf(buf, sizeof(buf), "%d", g_computer->GetMhz());
		g_input_mhz->value(buf);
		g_ask_speed->exec();
		mhz = atol(g_input_mhz->value());
		if(mhz <= 0 || mhz > 100) {
			// TODO: warn, we didn't take it!
			return;
		}
	}
	g_computer->SetMhz(mhz);

	g_1mhz_item->clear_value();
	g_2mhz_item->clear_value();
	g_4mhz_item->clear_value();
	g_6mhz_item->clear_value();
	g_8mhz_item->clear_value();
	g_10mhz_item->clear_value();
	g_other_mhz_item->clear_value();

	switch(mhz) {
	case 1:
		g_1mhz_item->set_value();
		break;

	case 2:
		g_2mhz_item->set_value();
		break;

	case 4:
		g_4mhz_item->set_value();
		break;

	case 6:
		g_6mhz_item->set_value();
		break;

	case 8:
		g_8mhz_item->set_value();
		break;

	case 10:
		g_10mhz_item->set_value();
		break;

	default:
		g_other_mhz_item->set_value();
		break;

	}
}

void SetCustomMhz(fltk::Button *, void *)
{
	g_ask_speed->hide();
}



void SetMonitorMhz(fltk::Item*, long mhz)
{
	g_refresh_rate = mhz;

	g_30mhz_monitor_item->clear_value();
	g_20mhz_monitor_item->clear_value();
	g_10mhz_monitor_item->clear_value();

	switch(mhz) {
	case 30:
		g_30mhz_monitor_item->set_value();
		break;

	case 20:
		g_20mhz_monitor_item->set_value();
		break;

	case 10:
		g_10mhz_monitor_item->set_value();
		break;

	}
}


void Apple::SetKey(unsigned char key)
{
	int		i;

	// do we assume that the key was already mapped?
	key |= 0x80;
	for(i = 0; i <= 16; ++i) {
		f_map[0xC][i] = key;
	}
}


void Apple::ResetMap(void)
{
	for(int i = 0; i < 16; ++i) {
		f_map[i] = f_memory + i * 4096;
	}
}



void Apple::GetVideoXY(unsigned short addr, int& x, int& y)
{
	// addresses go from 0x000 to 0x027 on the first line
	// 		from 0x080 to 0x0A7 on the second line
	// 		...
	// 		from 0x028 to 0x04F on the 8th line
	// 		from 0x0A8 to 0x0CF on the 9th line
	// 		...
	// 		from 0x050 to 0x077 on the 16th line
	// 		from 0x0D0 to 0x0F7 on the 17th line
	// 		...
	//
	// the following is used to avoid the % and / operators...
	int yp = addr & 0x7F;
	if(yp >= 0x50) {
		x = yp - 0x50;
		y = 16;
	}
	else if(yp >= 0x28) {
		x = yp - 0x28;
		y = 8;
	}
	else {
		x = yp;
		y = 0;
	}
	y += (addr & 0x03FF) >> 7;

	// hires has yet another trick to it!
	if(addr >= 0x2000) {
		if(addr >= 0x4000) {
			y += 24;
		}
		y = (y << 3) + ((addr >> 10) & 7);
	}
	else {
		if(addr >= 0x800) {
			y += 24;
		}
	}
}


void Apple::ChangedVideoXY(unsigned int *array, int x, int y)
{
	// mark the pixels which change
	int p = x + y * 40;
	array[p >> 5] |= 1 << (p & 0x1F);
}


// When a character changes we get a change in the
// text video and the lores video buffers
void Apple::DrawChar(unsigned short addr, unsigned char c)
{
	int			x, y, yp, h, w, page;
	unsigned int		*d;
	const unsigned char	*s;
	unsigned char		bits;

// First get the screen position in characters
	GetVideoXY(addr, x, y);
#if 0
	page = addr >= 0x800;
	y = yp + (page ? 24 : 0);
#else
	yp = y >= 24 ? y - 24 : y;
#endif

// for lores we add 0x800 to the address (that's like a virtual address!)
// we only draw one pixel per block; the rendering takes care of multiplying
// the single pixels as required! (i.e. 7 in width and 4 in height)
	// There is a bug (?) in gcc version 4.0
	//f_changed |= yp >= 20 ? CHANGED_FULL_LORES : CHANGED_LORES;
	if(yp >= 20) {
		f_changed |= CHANGED_FULL_LORES;
	}
	else {
		f_changed |= CHANGED_LORES;
	}
	ChangedVideoXY(f_lores_changed, x, y);
	d = f_lores_bitmaps + x + y * 40 * 2;
	d[0] = f_lores_colors[c & 0x0F];
	d[40] = f_lores_colors[c >> 4];

// text uses 8 lines and 7 pixels per line in our bitmaps
	// There is a bug (?) in gcc version 4.0
	//f_changed |= yp >= 20 ? CHANGED_FULL_TEXT : CHANGED_TEXT;
	if(yp >= 20) {
		f_changed |= CHANGED_FULL_TEXT;
	}
	else {
		f_changed |= CHANGED_TEXT;
	}
	ChangedVideoXY(f_text_changed, x, y);
	d = f_text_bitmaps + x * 7 + y * 40 * 7 * 8;

//printf("Draw '%c' ($%02X) at (%d, %d) [$%04X]\n", (c ^ 0x80) >= 0x20 && (c ^ 0x80) < 0x7F ? c ^ 0x80 : '?', c, x, y, addr);
//c = '@';

	s = f_characters + c * 8;
	for(h = 8; h > 0; --h, ++s, d += 39 * 7) {
		bits = *s;
		for(w = 7; w > 0; --w, ++d) {
			bits >>= 1;
			*d = f_text_colors[bits & 1];
		}
	}
}


void Apple::DrawPixels(unsigned short addr, unsigned char c0, unsigned char c1)
{
	int			x, y, w;
	unsigned int		*d;
	unsigned char		e0, e1;

// First get the screen position in pixels
	GetVideoXY(addr & 0xFFFE, x, y);
	assert((x & 1) == 0);

	// There is a bug (?) in gcc version 4.0
	//f_changed |= y >= 20 * 8 ? CHANGED_FULL_HIRES : CHANGED_HIRES;
	if(y >= 20 * 8) {
		f_changed |= CHANGED_FULL_HIRES;
	}
	else {
		f_changed |= CHANGED_HIRES;
	}
	ChangedVideoXY(f_hires_changed, x, y);

// Hires is rather complicated. It can be either black & white or color.
// When it uses colors, the first 6 bits of every two bytes would use
// bit 7 of the 1st byte to determine the high color bit selection.
// Similarly, bit 6 of the first byte and the 7 bits of the 2nd byte
// would be used with bit 7 of the 2nd byte to determine these colors.
// (note that means the f_hires_changed could use 'x / 2' intead of 'x')
	d = f_hires_bitmaps + x * 7 + y * 40 * 7;

	if(f_black_n_white) {
		for(w = 6; w > 0; --w, d++, c0 >>= 1, c1 >>= 1) {
			// TODO: should we have special B & W for hires?!
			d[0] = f_text_colors[c0 & 1];
			d[7] = f_text_colors[c1 & 1];
		}
	}
	else {
		// get the two bit 7
		e0 = (c0 >> 5) & 4;
		e1 = (c1 >> 5) & 4;
		// the middle bits come from both ends...
		// bit 7 of 2nd byte affects the color
		d[6] = d[7] = f_hires_colors[((c0 >> 6) & 1) | ((c1 & 1) << 1) | e1];
		c1 >>= 1;	// bit 1 used up already!
		for(w = 3; w > 0; --w, d += 2, c0 >>= 2, c1 >>= 2) {
			// note how the colors are duplicated!
			d[0] = d[1] = f_hires_colors[(c0 & 3) | e0];
			d[8] = d[9] = f_hires_colors[(c1 & 3) | e1];
		}
	}
}


typedef void (*draw_zoom_t)(Apple::rgb_t *dst, unsigned int color);


#define draw_zoom(z) static void Draw_Zoom##z(Apple::rgb_t *dst, unsigned int color) \
	{ \
		draw(red)   = (unsigned char) (color >> 16); \
		draw(green) = (unsigned char) (color >>  8); \
		draw(blue)  = (unsigned char) (color >>  0); \
	}

#define copy(x, c)	dst[x].c

#define draw(c)		copy(0, c)
draw_zoom(1)
#undef draw

#define	draw_sub(l, c)	copy((l) * 7 * 2 + 0, c) = copy((l) * 7 * 2 + 1, c)
#define	draw(c)		draw_sub(0, c) = draw_sub(40, c)
draw_zoom(2)
#undef draw
#undef draw_sub

#define	draw_sub(l, c)	copy((l) * 7 * 3 + 0, c) = copy((l) * 7 * 3 + 1, c) = copy((l) * 7 * 3 + 2, c)
#define	draw(c)		draw_sub(0, c) = draw_sub(40, c) = draw_sub(40 * 2, c)
draw_zoom(3)
#undef draw
#undef draw_sub

#define	draw_sub(l, c)	copy((l) * 7 * 4 + 0, c) = copy((l) * 7 * 4 + 1, c) = copy((l) * 7 * 4 + 2, c) = copy((l) * 7 * 4 + 3, c)
#define	draw(c)		draw_sub(0, c) = draw_sub(40, c) = draw_sub(40 * 2, c) = draw_sub(40 * 3, c)
draw_zoom(4)
#undef draw
#undef draw_sub

#define	draw_sub(l, c)	copy((l) * 7 * 5 + 0, c) = copy((l) * 7 * 5 + 1, c) = copy((l) * 7 * 5 + 2, c) = copy((l) * 7 * 5 + 3, c) = copy((l) * 7 * 5 + 4, c)
#define	draw(c)		draw_sub(0, c) = draw_sub(40, c) = draw_sub(40 * 2, c) = draw_sub(40 * 3, c) = draw_sub(40 * 4, c)
draw_zoom(5)
#undef draw
#undef draw_sub

#define	draw_sub(l, c)	copy((l) * 7 * 6 + 0, c) = copy((l) * 7 * 6 + 1, c) = copy((l) * 7 * 6 + 2, c) = copy((l) * 7 * 6 + 3, c) = copy((l) * 7 * 6 + 4, c) = copy((l) * 7 * 6 + 5, c)
#define	draw(c)		draw_sub(0, c) = draw_sub(40, c) = draw_sub(40 * 2, c) = draw_sub(40 * 3, c) = draw_sub(40 * 4, c) = draw_sub(40 * 5, c)
draw_zoom(6)
#undef draw
#undef draw_sub

#define	draw_sub(l, c)	copy((l) * 7 + 0, c) = copy((l) * 7 + 1, c) = copy((l) * 7 + 2, c) = copy((l) * 7 + 3, c) = copy((l) * 7 + 4, c) = copy((l) * 7 + 5, c) = copy((l) * 7 + 6, c)
#define	draw(c)		draw_sub(0, c) = draw_sub(40, c) = draw_sub(40 * 2, c) = draw_sub(40 * 3, c) = draw_sub(40 * 4, c) = draw_sub(40 * 5, c) = draw_sub(40 * 6, c)
draw_zoom(7)
#undef draw
#undef draw_sub

#define	draw_sub(l, c)	copy((l) * 7 * 8 + 0, c) = copy((l) * 7 * 8 + 1, c) = copy((l) * 7 * 8 + 2, c) = copy((l) * 7 * 8 + 3, c) = copy((l) * 7 * 8 + 4, c) = copy((l) * 7 * 8 + 5, c) = copy((l) * 7 * 8 + 6, c) = copy((l) * 7 * 8 + 7, c)
#define	draw(c)		draw_sub(0, c) = draw_sub(40, c) = draw_sub(40 * 2, c) = draw_sub(40 * 3, c) = draw_sub(40 * 4, c) = draw_sub(40 * 5, c) = draw_sub(40 * 6, c) = draw_sub(40 * 7, c)
draw_zoom(8)
#undef draw
#undef draw_sub


const draw_zoom_t tbl_draw[8] = {
	Draw_Zoom1,
	Draw_Zoom2,
	Draw_Zoom3,
	Draw_Zoom4,
	Draw_Zoom5,
	Draw_Zoom6,
	Draw_Zoom7,
	Draw_Zoom8
};


bool Apple::DrawText(rgb_t *dst, const unsigned int *src, unsigned int repeat, unsigned int *changed, int zoom, bool all)
{
	unsigned int		x, y, cnt, flg, line, pos;
	rgb_t			*d;
	const unsigned int	*s;
	draw_zoom_t		draw;
	bool			result;

	draw = tbl_draw[zoom - 1];

	result = false;

	line = 40;
	for(pos = 0; pos < repeat; ++pos) {
		flg = all ? (unsigned int) -1 : changed[pos];
		changed[pos] = 0;
		if(flg) {
			cnt = 32;
			do {
				if(flg & 1) {
					result = true;
					s = src;
					d = dst;
					y = 8;
					do {
						x = 7;
						do {
							(*draw)(d, *s);
							s++;
							d += zoom;
							x--;
						} while(x > 0);
						s += 40 * 7 - 7;
						d += 40 * 7 * zoom * zoom - 7 * zoom;
						y--;
					} while(y > 0);
				}
				src += 7;
				dst += 7 * zoom;
				line--;
				if(line == 0) {
					line = 40;
					src += 40 * 7 * 7;
					dst += (40 * 7) * zoom * (8 * zoom - 1);
				}
				flg >>= 1;
				cnt--;
			} while(cnt > 0);
		}
		else {
			src += 7 * 32;
			dst += (7 * 32) * zoom;
			line -= 32;
			if((int) line <= 0) {
				line += 40;
				src += 40 * 7 * 7;
				dst += (40 * 7) * zoom * (8 * zoom - 1);
			}
		}
	}

	return result;
}



bool Apple::DrawLoRes(rgb_t *dst, const unsigned int *src, unsigned int repeat, unsigned int *changed, int zoom, bool all)
{
	unsigned int		x, y, cnt, flg, line, pos, color;
	rgb_t			*d;
	const unsigned int	*s;
	draw_zoom_t		draw;
	bool			result;

	draw = tbl_draw[zoom - 1];

	result = false;

	line = 40;
	for(pos = 0; pos < repeat; ++pos) {
		flg = all ? (unsigned int) -1 : changed[pos];
		changed[pos] = 0;
		if(flg) {
			cnt = 32;
			do {
				if(flg & 1) {
					result = true;
					d = dst;
					color = src[0];
					y = 4;
					do {
						x = 7;
						do {
							(*draw)(d, color);
							d += zoom;
							x--;
						} while(x > 0);
						d += (40 * 7) * zoom * zoom - 7 * zoom;
						y--;
					} while(y > 0);
					color = src[40];
					y = 4;
					do {
						x = 7;
						do {
							(*draw)(d, color);
							d += zoom;
							x--;
						} while(x > 0);
						d += (40 * 7) * zoom * zoom - 7 * zoom;
						y--;
					} while(y > 0);
				}
				src += 1;
				dst += 7 * zoom;
				line--;
				if(line == 0) {
					line = 40;
					src += 40;
					dst += (40 * 7) * zoom * (8 * zoom - 1);
				}
				flg >>= 1;
				cnt--;
			} while(cnt > 0);
		}
		else {
			src += 32;
			dst += (32 * 7) * zoom;
			line -= 32;
			if((int) line <= 0) {
				line += 40;
				src += 40;
				dst += (40 * 7) * zoom * (8 * zoom - 1);
			}
		}
	}

	return result;
}



bool Apple::DrawHiRes(rgb_t *dst, const unsigned int *src, unsigned int repeat, unsigned int *changed, int zoom, bool all)
{
	unsigned int		x, cnt, flg, pos, line;
	draw_zoom_t		draw;
	bool			result;

	draw = tbl_draw[zoom - 1];

	result = false;

	line = 40;
	for(pos = 0; pos < repeat; ++pos) {
		flg = all ? (unsigned int) -1 : changed[pos];
		changed[pos] = 0;
		if(flg) {
			cnt = 32;
			do {
				// only even bit set for every two pixels
				if(flg & 3) {
					result = true;
					x = 14;
					do {
						(*draw)(dst, *src);
						src++;
						dst += zoom;
						x--;
					} while(x > 0);
				}
				else {
					src += 14;
					dst += 14 * zoom;
				}
				line -= 2;
				if(line == 0) {
					line = 40;
					dst += (40 * 7) * zoom * (zoom - 1);
				}
				flg >>= 2;
				cnt -= 2;
			} while(cnt > 0);
		}
		else {
			src += 32 * 7;
			dst += (32 * 7) * zoom;
			line -= 32;
			if((int) line <= 0) {
				line += 40;
				dst += (40 * 7) * zoom * (zoom - 1);
			}
		}
	}

	return result;
}




bool Apple::CreateRGBImage(rgb_t *rgb, int zoom, bool all)
{
	// here we first want to fill the first 20 lines
	// then the last 4 lines -- thus if we are in mixed
	// mode we can render exactly what's necessary

	unsigned int		page, mode, res, select_mode, changed;
	draw_zoom_t		draw;
	bool			result;

	// the following should be thread protected
	changed = f_changed;
	f_changed = 0;

	// whether we've drawn something (something changed on the screen)
	result = false;

	assert(zoom >= 1 && zoom <= 8);

	select_mode = (f_user_video_mode & VIDEO_MODE_SELECT_MASK) == VIDEO_MODE_PROGRAM_SELECT ? f_video_mode : f_user_video_mode;

//printf("RBG from %02X (%02X vs %02X)\n", select_mode, f_user_video_mode, f_video_mode);

	page = select_mode & VIDEO_MODE_PAGE_MASK;
	mode = select_mode & VIDEO_MODE_MASK;
	if(mode == (VIDEO_MODE_TEXT | VIDEO_MODE_MIXED)) {
		// ignore MIXED whenever we don't have GRAPHICS
		mode = VIDEO_MODE_TEXT;
	}
	res = select_mode & VIDEO_MODE_RES_MASK;

	if(mode == VIDEO_MODE_TEXT) {
		// text
		//
		// 20 lines of 40 characters (and 32 bits per flg)
		// (note that 40 * 20 is a multiple of 32)
		if(changed & CHANGED_TEXT) {
			result |= DrawText(
				rgb,
				f_text_bitmaps + page * (40 * 7 * 24 * 8),
				40 * 20 / 32,
				f_text_changed + page * (40 * 24 / 32),
				zoom, all);
		}
	}
	else {
		// graphics or mixed, we have graphics here
		if(res == VIDEO_MODE_LORES) {
			// lores
			if(changed & CHANGED_LORES) {
				result |= DrawLoRes(
					rgb,
					f_lores_bitmaps + page * (40 * 48),
					40 * 20 / 32,
					f_lores_changed + page * (40 * 48 / 32),
					zoom, all);
			}
		}
		else {
			// hires
			if(changed & CHANGED_HIRES) {
				result |= DrawHiRes(
					rgb,
					f_hires_bitmaps + page * (40 * 7 * 24 * 8),
					40 * 160 / 32,
					f_hires_changed + page * (40 * 192 / 32),
					zoom, all);
			}
		}
	}

	if(mode == VIDEO_MODE_GRAPHICS) {
		// graphics or mixed, we have graphics here
		if(res == VIDEO_MODE_LORES) {
			// lores
			if(changed & CHANGED_FULL_LORES) {
				result |= DrawLoRes(
					rgb + 40 * 7 * zoom * 20 * 8 * zoom,
					f_lores_bitmaps + page * (40 * 48) + 40 * 40,
					40 * 20 / 32,
					f_lores_changed + page * (40 * 48 / 32) + 40 * 40 / 32,
					zoom, all);
			}
		}
		else {
			// hires
			if(changed & CHANGED_FULL_HIRES) {
				result |= DrawHiRes(
					rgb + 40 * 7 * zoom * 20 * 8 * zoom,
					f_hires_bitmaps + page * (40 * 7 * 24 * 8) + 40 * 7 * 20 * 8,
					40 * 32 / 32,
					f_hires_changed + page * (40 * 192 / 32) + 40 * 160 / 32,
					zoom, all);
			}
		}
	}
	else {
		// text or mixed
		//
		// 4 lines of 40 characters (and 32 bits per flg)
		// (note that 4 * 20 is a multiple of 32)
		if(changed & CHANGED_FULL_TEXT) {
			result |= DrawText(
				rgb + 40 * 7 * zoom * 20 * 8 * zoom,
				f_text_bitmaps + page * (40 * 7 * 24 * 8) + 40 * 7 * 20 * 8,
				40 * 4 / 32,
				f_text_changed + page * (40 * 24 / 32) + 40 * 20 / 32,
				zoom, all);
		}
	}

	return result;
}



Apple		*g_apple;


void QuitApp(fltk::Item*, void*)
{
	g_window->hide();
}

void SwapStats(fltk::Item *item, void*)
{
	if(item->value()) {
		g_window->resize(560, 509);
	}
	else {
		g_window->resize(560, 409);
	}
}

void SelectOutputMode(fltk::Item *item, void*)
{
	unsigned int	video_mode;

	video_mode = 0;

	// Program select, text, lores or hires
	if(item == g_program_select_item) {
program:
		video_mode |= Apple::VIDEO_MODE_PROGRAM_SELECT;
		g_program_select_item->set_value();
		g_text_select_item->clear_value();
		g_lores_select_item->clear_value();
		g_hires_select_item->clear_value();
	}
	else if(item == g_text_select_item) {
text:
		video_mode |= Apple::VIDEO_MODE_USER_SELECT;
		g_text_select_item->set_value();
		g_program_select_item->clear_value();
		g_lores_select_item->clear_value();
		g_hires_select_item->clear_value();
	}
	else if(item == g_lores_select_item) {
lores:
		video_mode |= Apple::VIDEO_MODE_USER_SELECT | Apple::VIDEO_MODE_GRAPHICS;
		g_lores_select_item->set_value();
		g_program_select_item->clear_value();
		g_text_select_item->clear_value();
		g_hires_select_item->clear_value();
	}
	else if(item == g_hires_select_item) {
hires:
		video_mode |= Apple::VIDEO_MODE_USER_SELECT | Apple::VIDEO_MODE_GRAPHICS | Apple::VIDEO_MODE_HIRES;
		g_hires_select_item->set_value();
		g_program_select_item->clear_value();
		g_text_select_item->clear_value();
		g_lores_select_item->clear_value();
	}
	// if none of the program/text/lores/hires check which one is
	// currently selected instead
	else if(g_program_select_item->value()) {
		goto program;
	}
	else if(g_text_select_item->value()) {
		goto text;
	}
	else if(g_lores_select_item->value()) {
		goto lores;
	}
	else if(g_hires_select_item->value()) {
		goto hires;
	}

	// Mixed (ignored when text is selected)
	if(g_mixed_select_item->value() && (video_mode & Apple::VIDEO_MODE_MASK) == Apple::VIDEO_MODE_GRAPHICS) {
		video_mode |= Apple::VIDEO_MODE_MIXED;
	}

	// Page 0 or 1
	if(g_screen_select_item->value()) {
		video_mode |= Apple::VIDEO_MODE_PAGE1;
		g_screen_select_item->label("Screen 1");
	}
	else {
		video_mode |= Apple::VIDEO_MODE_PAGE0;
		g_screen_select_item->label("Screen 0");
	}

	g_apple->SelectUserVideo(video_mode);
}

void SwapTrace(fltk::Item *item, void*)
{
	g_computer->SetTrace(item->value());
}


void LoadBinaryFile(fltk::Item *, void *)
{
	if(g_file_chooser == 0) {
		//g_file_chooser = new fltk::FileChooser(".", "Object Files (*.{obj,o})\tAll Files (*.*)", fltk::FileChooser::SINGLE, "Select a pre-compiled 6502 binary file");
		g_file_chooser = new fltk::FileChooser(".", "*.o", fltk::FileChooser::SINGLE, "Select a pre-compiled 6502 binary file");
	}
	g_file_chooser->exec();
	if(g_file_chooser->count() == 0) {
		// cancelled
		return;
	}
	const char *filename = g_file_chooser->value();
	g_ask_address->exec();
	const char *addr = g_input_address->value();

	char *end;
	long ptr = strtol(addr, &end, 0);
	// TODO: test validity + error message

	//fprintf(stderr, "Load [%s] at [%d]\n", file, ptr);

	g_apple->LoadFile(filename, ptr);
}

void LoadBinaryAddress(fltk::Button *, void *)
{
	g_ask_address->hide();
}


void SaveBinaryFile(fltk::Item *, void *)
{
	if(g_file_chooser == 0) {
		//g_file_chooser = new fltk::FileChooser(".", "Object Files (*.{obj,o})\tAll Files (*.*)", fltk::FileChooser::SINGLE, "Select a pre-compiled 6502 binary file");
		g_file_chooser = new fltk::FileChooser(".", "*.o", fltk::FileChooser::SINGLE, "Select a pre-compiled 6502 binary file");
	}

	g_ask_address_and_size->exec();
	const char *addr = g_save_address->value();
	const char *size = g_save_size->value();

	g_file_chooser->exec();
	if(g_file_chooser->count() == 0) {
		// cancelled
		return;
	}
	const char *filename = g_file_chooser->value();

	char *end;
	long ptr = strtol(addr, &end, 0);
	long len = strtol(size, &end, 0);
	// TODO: test validity + error message

	g_apple->SaveFile(filename, ptr, len);
}

void SaveBinaryAddressAndSize(fltk::Button *, void *)
{
	g_ask_address_and_size->hide();
}


void LoadBasicProgram(fltk::Item *, void *)
{
	if(g_file_chooser == 0) {
		//g_file_chooser = new fltk::FileChooser(".", "Object Files (*.{obj,o})\tAll Files (*.*)", fltk::FileChooser::SINGLE, "Select a pre-compiled 6502 binary file");
		g_file_chooser = new fltk::FileChooser(".", "*.o", fltk::FileChooser::SINGLE, "Select a pre-compiled 6502 binary file");
	}

	g_file_chooser->exec();
	if(g_file_chooser->count() == 0) {
		// cancelled
		return;
	}
	const char *filename = g_file_chooser->value();

	g_apple->LoadFile(filename, 0, true);
}


void SaveBasicProgram(fltk::Item *, void *)
{
	if(g_file_chooser == 0) {
		//g_file_chooser = new fltk::FileChooser(".", "Object Files (*.{obj,o})\tAll Files (*.*)", fltk::FileChooser::SINGLE, "Select a pre-compiled 6502 binary file");
		g_file_chooser = new fltk::FileChooser(".", "*.o", fltk::FileChooser::SINGLE, "Select a pre-compiled 6502 binary file");
	}

	g_file_chooser->exec();
	if(g_file_chooser->count() == 0) {
		// cancelled
		return;
	}
	const char *filename = g_file_chooser->value();

	g_apple->SaveFile(filename, 0, 0, true);
}



void ResetNow(fltk::Item *, void *)
{
	g_computer->Reset();
}


void About(fltk::Item *, void *)
{
	g_about_name->value(PACKAGE_NAME);
	g_about_version->value("Version: " PACKAGE_VERSION);
	g_about_copyright->value("Copyright (c) 2005-2006 by Alexis Wilke");
	g_about_author->value("Written by Alexis Wilke");

	g_about_name->position(0);
	g_about_version->position(0);
	g_about_copyright->position(0);
	g_about_author->position(0);

	g_about->show_inside(g_window);
}


void CloseAbout(fltk::Button *, void *)
{
	g_about->hide();
}


void License(fltk::Item *, void *)
{
	g_license_info->position(0);
	g_license->show_inside(g_window);
}


void CloseLicense(fltk::Button *, void *)
{
	g_license->hide();
}


void Copyright(fltk::Item *, void *)
{
	g_copyright_info->position(0);
	g_copyright->show_inside(g_window);
}


void CloseCopyright(fltk::Button *, void *)
{
	g_copyright->hide();
}


void Keyboard(unsigned int key)
{
	g_apple->SetKey(key);
}



void CheckScreen(void *app)
{
	Apple		*apple;

	apple = reinterpret_cast<Apple *>(app);

	if(apple->CreateRGBImage(g_screen_data, g_zoom, false)) {

#if 0
// for test purposes, save the resulting image in a targa file named test.tga
static bool done = false;
if(!done) {
done = true;
FILE *f = fopen("test.tga", "w");
char header[18];
header[ 0] = 0;		// no identifier
header[ 1] = 0;		// no colormap
header[ 2] = 2;		// uncompressed RGB image
header[ 3] = 0;		// colormap stuff
header[ 4] = 0;		// colormap stuff
header[ 5] = 0;		// colormap stuff
header[ 6] = 0;		// colormap stuff
header[ 7] = 0;		// colormap stuff
header[ 8] = 0;		// origin stuff
header[ 9] = 0;		// origin stuff
header[10] = 0;		// origin stuff
header[11] = 0;		// origin stuff
int zoom = g_zoom;
header[12] = 40 * 7 * zoom;	// width low
header[13] = (40 * 7 * zoom) >> 8;	// width high
header[14] = 24 * 8 * zoom;	// height low
header[15] = (24 * 8 * zoom) >> 8;	// height high
header[16] = 24;	// bits per pixel
header[17] = 0x20;	// flags: "up side down" (i.e. "normal" in my world!)
fwrite(&header, 1, 18, f);
fwrite(g_screen_data, 1, 40 * 7 * g_zoom * 24 * 8 * g_zoom * 3, f);
fclose(f);
}
#endif

		// something changed, let's show it to the user
		// TODO: should we create a set of regions to only
		//	 draw what needs to be refreshed?!
		g_screen->destroy_cache();
		g_screen_widget->redraw();
		/*
		fltk::Rectangle rect(0, 25, 40 * 7 * g_zoom, 24 * 8 * g_zoom);
		g_screen->draw(rect, 0, 0);
		*/
	}

	// time out 30/20/10 times per second
	fltk::repeat_timeout(1.0f / g_refresh_rate, CheckScreen, apple);
}



static void usage(const char *argv0)
{
	const char *basename = strrchr(argv0, '/');
	if(basename == 0) {
		basename = argv0;
	}
	else {
		basename++;
	}

	fflush(stdout);
	fprintf(stderr, "\n");
	fprintf(stderr, "%s v%s\n", basename, PACKAGE_VERSION);
	fprintf(stderr, "\n");
	fprintf(stderr, "Usage: %s [-opt] [rom] [slot6]\n", basename);
	fprintf(stderr, "  -c <cpu>    defines the CPU to use (6502, 65C02 or 65C02S)\n");
	fprintf(stderr, "  -i <id>     defines the ID to use (65C02)\n");
	fprintf(stderr, "  -h          prints out this help screen\n");
	fprintf(stderr, "  -t          trace as the processor runs\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	emulator_8bits::emu65xx::id_t		id;
	emulator_8bits::emu65xx::processor_t	cpu;
	bool					trace;
	int					i, j, l;

	id = emulator_8bits::emu65xx::ID_65C02;
	cpu = emulator_8bits::emu65xx::CPU_65C02S;
	trace = false;

	i = 1;
	while(i < argc) {
		if(argv[i][0] == '-') {
			l = strlen(argv[i]);
			j = 1;
			while(j < l) {
				switch(argv[i][j]) {
				case 'c':
					j++;
					if(j >= l) {
						i++;
						if(i >= argc) {
fprintf(stderr, "ERROR: -c is expecting to be followed by a CPU name\n");
							usage(argv[0]);
						}
						j = 0;
						if(argv[i][0] == '-') {
fprintf(stderr, "ERROR: -c is expecting to be followed by a CPU name\n");
							usage(argv[0]);
						}
					}
					if(strcmp(argv[i] + j, "6502") == 0) {
						cpu = emulator_8bits::emu65xx::CPU_6502;
					}
					else if(strcmp(argv[i] + j, "65C02") == 0) {
						cpu = emulator_8bits::emu65xx::CPU_65C02;
					}
					else if(strcmp(argv[i] + j, "65C02S") == 0) {
						cpu = emulator_8bits::emu65xx::CPU_65C02S;
					}
					else {
fprintf(stderr, "ERROR: the CPU \"%s\" is not supported.\n", argv[i] + j);
							usage(argv[0]);
					}
					j = l;
					break;

				case 'i':
					j++;
					if(j >= l) {
						i++;
						if(i >= argc) {
fprintf(stderr, "ERROR: -i is expecting to be followed by a ID name\n");
							usage(argv[0]);
						}
						j = 0;
						if(argv[i][0] == '-') {
fprintf(stderr, "ERROR: -i is expecting to be followed by a ID name\n");
							usage(argv[0]);
						}
					}
					if(strcmp(argv[i] + j, "65C02") == 0) {
						id = emulator_8bits::emu65xx::ID_65C02;
					}
					else {
fprintf(stderr, "ERROR: the ID \"%s\" is not supported.\n", argv[i] + j);
						usage(argv[0]);
					}
					j = l;
					break;

				case 'h':
					usage(argv[0]);
					break;

				case 't':
					trace = true;
					break;

				}
				j++;
			}
		}
		i++;
	}

	// create the Apple with a new operator; in the BSS section
	// it could break your executable on some systems
	g_apple = new Apple;

	g_computer = emulator_8bits::emu65xx::Create(id, cpu, *g_apple);
	g_computer->SetTrace(trace);

	g_apple->LoadROM("apple_II.rom", 0xD000, 0x3000);

	g_apple->PatchROM(0xFCA8, 0x60); // use if you want to debug the trace output...

	//g_apple->LoadROM("slot6.rom", 0xC600, 0x100);	// until the disk works...
	g_apple->LoadCharacters("characters.rom");

	g_computer->Reset();

	g_window = make_window();

	g_about = make_about();
	g_license = make_license();
	g_copyright = make_copyright();
	g_ask_address = ask_address();
	g_ask_speed = ask_speed();
	g_ask_address_and_size = ask_address_and_size();

	g_window->resize(560, 409);	// hide the stats

	if(trace) {
		g_trace_item->set_value();
	}

	g_window->show();

	// The screen data buffer size is defined as follow
	// 40 columns of 7 pixels zoomed up to x 8
	// 24 rows of 8 pixels zoomed up to x 8
	g_zoom = 2;	// start with a zoom of 2
	g_screen_data = new Apple::rgb_t[40 * 7 * 8 * 24 * 8 * 8];

#if 1
//_DEBUG
	memset(g_screen_data, 0xAA, 40 * 7 * 8 * 24 * 8 * 8 * sizeof(Apple::rgb_t));
	{
		Apple::rgb_t *d = g_screen_data;
		for(int y = 0; y < 24 * 8 * 8; y++) {
			for(int x = 0; x < 40 * 7 * 8; x++, d++) {
				if((x ^ y) & 1) {
					if((y & 4) == 0) {
						d->red   = 255;
						d->green = 255;
						d->blue  = 255;
					}
					else {
						d->red   =   0;
						d->green = 255;
						d->blue  = 255;
					}
				}
				else {
					d->red   =   0;
					d->green =   0;
					d->blue  =   0;
				}
			}
		}
	}
#endif

	g_screen = new fltk::rgbImage((unsigned char *) g_screen_data, 40 * 7 * g_zoom, 24 * 8 * g_zoom);
	g_screen_widget->image(g_screen);
	g_screen_widget->zoom(g_zoom);

	// time out 30/20/10 times per second
	fltk::add_timeout(1.0f / g_refresh_rate, CheckScreen, g_apple);

	fltk::run();

	g_computer->Release();
}

