/* emu_addr.h -- written by Alexis WILKE (c) 2005-2006 */

/**********************************************************************
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**********************************************************************/


#include	"emu.h"

namespace emulator_8bits
{


// Fetch Address functions
bool emu65C02::FetchAddress_NONE(void)
{
	return false;
}

bool emu65C02::FetchAddress_IMMEDIATE(void)
{
	f_addr = f_regs.f_pc;
	f_regs.f_pc++;
	return false;
}

bool emu65C02::FetchAddress_PZ(void)
{
	f_addr = Load(f_regs.f_pc);
	f_regs.f_pc++;
	return false;
}

bool emu65C02::FetchAddress_PZX(void)
{
	f_addr = (Load(f_regs.f_pc) + f_regs.f_x) & 255;
	f_regs.f_pc++;
	return false;
}

bool emu65C02::FetchAddress_PZY(void)
{
	f_addr = (Load(f_regs.f_pc) + f_regs.f_y) & 255;
	f_regs.f_pc++;
	return false;
}

bool emu65C02::FetchAddress_INDPZ(void)
{
	f_addr = LoadWordRound(Load(f_regs.f_pc));
	f_regs.f_pc++;
	return false;
}

bool emu65C02::FetchAddress_INDPZX(void)
{
	f_addr = LoadWordRound((Load(f_regs.f_pc) + f_regs.f_x) & 255);
	f_regs.f_pc++;
	return false;
}

bool emu65C02::FetchAddress_INDPZY(void)
{
	unsigned short	addr;

	addr = LoadWordRound(Load(f_regs.f_pc));
	f_addr = addr + f_regs.f_y;
	f_regs.f_pc++;

	return (addr >> 8) != (f_addr >> 8);
}

bool emu65C02::FetchAddress_ABS(void)
{
	f_addr = LoadWord(f_regs.f_pc);
	f_regs.f_pc += 2;
	return false;
}

bool emu65C02::FetchAddress_ABSX(void)
{
	unsigned short	addr;

	addr = LoadWord(f_regs.f_pc);
	f_addr = addr + f_regs.f_x;
	f_regs.f_pc += 2;

	return (addr >> 8) != (f_addr >> 8);
}

bool emu65C02::FetchAddress_ABSY(void)
{
	unsigned short	addr;

	addr = LoadWord(f_regs.f_pc);
	f_addr = LoadWord(f_regs.f_pc) + f_regs.f_y;
	f_regs.f_pc += 2;

	return (addr >> 8) != (f_addr >> 8);
}


bool emu65C02::FetchAddress_INDABS6502(void)
{
	// there was a bug on older processor with absolute indirect
	f_addr = LoadWordRound(LoadWord(f_regs.f_pc));
	f_regs.f_pc += 2;
	return false;
}


bool emu65C02::FetchAddress_INDABS(void)
{
	f_addr = LoadWord(LoadWord(f_regs.f_pc));
	f_regs.f_pc += 2;
	return false;
}


bool emu65C02::FetchAddress_INDABSX(void)
{
	f_addr = LoadWord(LoadWord(f_regs.f_pc) + f_regs.f_x);
	f_regs.f_pc += 2;
	return false;
}


bool emu65C02::FetchAddress_RELATIVE(void)
{
	f_addr = (signed char) Load(f_regs.f_pc) + f_regs.f_pc + 1;
	f_regs.f_pc++;
	return false;
}



};		// emulator_8bits
// vim: ts=8
