/* emu_core.h -- written by Alexis WILKE (c) 2005-2006 */

/**********************************************************************
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**********************************************************************/


#include	"emu.h"

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>


namespace emulator_8bits
{

// define the instructions; this includes which processor runs what
#define	I(x, n, a, l, s, p)	{ 0X##x, #n,    &emu65C02::inst_##n, ADDR_##a,  l, s, p }
#define	U(x)			{ 0X##x, "???", &emu65C02::inst_NOP, ADDR_NONE, 0, 0, 0 }

//#define	P1(n)		(1 << (CPU_##n))
//#define	P2(n,m)		(P1(n)|P1(m))
//#define	P3(n,m,o)	(P1(n)|P1(m)|P1(o))

#define	P1(n)		{ 0, 0, (n) }
#define	P2(n,m)		{ 0, (n), (m) }
#define	P3(n,m,o)	{ (n), (m), (o) }


const emu65C02::inst_t emu65C02::g_inst[256] =
{
//	  x	inst	address		load	store	processor
	I(00,	BRK,	IMMEDIATE,	0,	0,	P3(0x07,0x07,0x07)),	// BRK
	I(01,	ORA,	INDPZX,		1,	0,	P3(0x06,0x06,0x06)),	// ORA (PZ,X)
	U(02),
	U(03),
	I(04,	TSB,	PZ,		1,	1,	P2(0x05,0x05)),		// TSB PZ
	I(05,	ORA,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// ORA PZ
	I(06,	ASL,	PZ,		1,	1,	P3(0x05,0x05,0x05)),	// ASL PZ
	I(07,	RMB,	PZ,		1,	1,	P1(0x05)),		// RMB0 PZ
	I(08,	PHP,	NONE,		0,	0,	P3(0x03,0x03,0x03)),	// PHP
	I(09,	ORA,	IMMEDIATE,	1,	0,	P3(0x02,0x02,0x02)),	// ORA #IMM
	I(0A,	ASLA,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// ASL ACC
	U(0B),
	I(0C,	TSB,	ABS,		1,	1,	P2(0x06,0x06)),		// TSB ABS
	I(0D,	ORA,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// ORA ABS
	I(0E,	ASL,	ABS,		1,	1,	P3(0x06,0x06,0x06)),	// ASL ABS
	I(0F,	BBR,	PZ,		1,	0,	P1(0xC5)),		// BBR0 PZ,RELATIVE

//	  x	inst	address		load	store
	I(10,	BPL,	RELATIVE,	0,	0,	P3(0xC2,0xC2,0xC2)),	// BPL RELATIVE
	I(11,	ORA,	INDPZY,		1,	0,	P3(0x85,0x85,0x85)),	// ORA (PZ),Y
	I(12,	ORA,	INDPZ,		1,	0,	P2(0x05,0x05)),		// ORA (PZ)
	U(13),
	I(14,	TRB,	PZ,		1,	1,	P2(0x05,0x05)),		// TRB PZ
	I(15,	ORA,	PZX,		1,	0,	P3(0x04,0x04,0x04)),	// ORA PZ,X
	I(16,	ASL,	PZX,		1,	1,	P3(0x06,0x06,0x06)),	// ASL PZ,X
	I(17,	RMB,	PZ,		1,	1,	P1(0x05)),		// RMB1 PZ
	I(18,	CLC,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// CLC
	I(19,	ORA,	ABSY,		1,	0,	P3(0x84,0x84,0x84)),	// ORA ABS,Y
	I(1A,	INCA,	NONE,		0,	0,	P2(0x02,0x02)),		// INC ACC
	U(1B),
	I(1C,	TRB,	ABS,		1,	1,	P2(0x06,0x06)),		// TRB ABS
	I(1D,	ORA,	ABSX,		1,	0,	P3(0x84,0x84,0x84)),	// ORA ABS,X
	I(1E,	ASL,	ABSX,		1,	1,	P3(0x07,0x07,0x07)),	// ASL ABS,X
	I(1F,	BBR,	PZ,		1,	0,	P1(0xC5)),		// BBR1 PZ,RELATIVE

//	  x	inst	address		load	store
	I(20,	JSR,	ABS,		0,	0,	P3(0x06,0x06,0x06)),	// JSR ABS
	I(21,	AND,	INDPZX,		1,	0,	P3(0x06,0x06,0x06)),	// AND (PZ,X)
	U(22),
	U(23),
	I(24,	BIT,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// BIT PZ
	I(25,	AND,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// AND PZ
	I(26,	ROL,	PZ,		1,	1,	P3(0x05,0x05,0x05)),	// ROL PZ
	I(27,	RMB,	PZ,		1,	1,	P1(0x05)),		// RMB2 PZ
	I(28,	PLP,	NONE,		0,	0,	P3(0x04,0x04,0x04)),	// PLP
	I(29,	AND,	IMMEDIATE,	1,	0,	P3(0x02,0x02,0x02)),	// AND #IMM
	I(2A,	ROLA,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// ROL ACC
	U(2B),
	I(2C,	BIT,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// BIT ABS
	I(2D,	AND,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// AND ABS
	I(2E,	ROL,	ABS,		1,	1,	P3(0x06,0x06,0x06)),	// ROL ABS
	I(2F,	BBR,	PZ,		1,	0,	P1(0xC5)),		// BBR2 PZ,RELATIVE

//	  x	inst	address		load	store
	I(30,	BMI,	RELATIVE,	0,	0,	P3(0xC2,0xC2,0xC2)),	// BMI RELATIVE
	I(31,	AND,	INDPZY,		1,	0,	P3(0x85,0x85,0x85)),	// AND (PZ),Y
	I(32,	AND,	INDPZ,		1,	0,	P2(0x05,0x05)),		// AND (PZ)
	U(33),
	I(34,	BIT,	PZX,		1,	0,	P2(0x03,0x03)),		// BIT PZ,X
	I(35,	AND,	PZX,		1,	0,	P3(0x04,0x04,0x04)),	// AND PZ,X
	I(36,	ROL,	PZX,		1,	1,	P3(0x06,0x06,0x06)),	// ROL PZ,X
	I(37,	RMB,	PZ,		1,	1,	P1(0x05)),		// RMB3 PZ
	I(38,	SEC,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// SEC
	I(39,	AND,	ABSY,		1,	0,	P3(0x84,0x84,0x84)),	// AND ABS,Y
	I(3A,	DECA,	NONE,		0,	0,	P2(0x02,0x02)),		// DEC ACC
	U(3B),
	I(3C,	BIT,	ABSX,		1,	0,	P2(0x04,0x04)),		// BIT ABS,X
	I(3D,	AND,	ABSX,		1,	0,	P3(0x84,0x84,0x84)),	// AND ABS,X
	I(3E,	ROL,	ABSX,		1,	1,	P3(0x07,0x07,0x07)),	// ROL ABS,X
	I(3F,	BBR,	PZ,		1,	0,	P1(0xC5)),		// BBR3 PZ,RELATIVE

//	  x	inst	address		load	store
	I(40,	RTI,	NONE,		0,	0,	P3(0x06,0x06,0x06)),	// RTI
	I(41,	EOR,	INDPZX,		1,	0,	P3(0x06,0x06,0x06)),	// EOR (PZ,X)
	U(42),
	U(43),
	U(44),
	I(45,	EOR,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// EOR PZ
	I(46,	LSR,	PZ,		1,	1,	P3(0x05,0x05,0x05)),	// LSR PZ
	I(47,	RMB,	PZ,		1,	1,	P1(0x05)),		// RMB4 PZ
	I(48,	PHA,	NONE,		0,	0,	P3(0x03,0x03,0x03)),	// PHA
	I(49,	EOR,	IMMEDIATE,	1,	0,	P3(0x02,0x02,0x02)),	// EOR #IMM
	I(4A,	LSRA,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// LSR ACC
	U(4B),
	I(4C,	JMP,	ABS,		0,	0,	P3(0x03,0x03,0x03)),	// JMP ABS
	I(4D,	EOR,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// EOR ABS
	I(4E,	LSR,	ABS,		1,	1,	P3(0x06,0x06,0x06)),	// LSR ABS
	I(4F,	BBR,	PZ,		1,	0,	P1(0xC5)),		// BBR4 PZ,RELATIVE

//	  x	inst	address		load	store
	I(50,	BVC,	RELATIVE,	0,	0,	P3(0xC2,0xC2,0xC2)),	// BVC RELATIVE
	I(51,	EOR,	INDPZY,		1,	0,	P3(0x85,0x85,0x85)),	// EOR (PZ),Y
	I(52,	EOR,	INDPZ,		1,	0,	P2(0x05,0x05)),		// EOR (PZ)
	U(53),
	U(54),
	I(55,	EOR,	PZX,		1,	0,	P3(0x04,0x04,0x04)),	// EOR PZ,X
	I(56,	LSR,	PZX,		1,	1,	P3(0x06,0x06,0x06)),	// LSR PZ,X
	I(57,	RMB,	PZ,		1,	1,	P1(0x05)),		// RMB5 PZ
	I(58,	CLI,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// CLI
	I(59,	EOR,	ABSY,		1,	0,	P3(0x84,0x84,0x84)),	// EOR ABS,Y
	I(5A,	PHY,	NONE,		0,	0,	P2(0x03,0x03)),		// PHY
	U(5B),
	U(5C),
	I(5D,	EOR,	ABSX,		1,	0,	P3(0x84,0x84,0x84)),	// EOR ABS,X
	I(5E,	LSR,	ABSX,		1,	1,	P3(0x07,0x07,0x07)),	// LSR ABS,X
	I(5F,	BBR,	PZ,		1,	0,	P1(0xC5)),		// BBR5 PZ,RELATIVE

//	  x	inst	address		load	store
	I(60,	RTS,	NONE,		0,	0,	P3(0x06,0x06,0x06)),	// RTS
	I(61,	ADC,	INDPZX,		1,	0,	P3(0x26,0x26,0x26)),	// ADC (PZ,X)
	U(62),
	U(63),
	I(64,	STZ,	PZ,		0,	1,	P2(0x03,0x03)),		// STZ PZ
	I(65,	ADC,	PZ,		1,	0,	P3(0x23,0x23,0x23)),	// ADC PZ
	I(66,	ROR,	PZ,		1,	1,	P3(0x05,0x05,0x05)),	// ROR PZ
	I(67,	RMB,	PZ,		1,	1,	P1(0x05)),		// RMB6 PZ
	I(68,	PLA,	NONE,		0,	0,	P3(0x04,0x04,0x04)),	// PLA
	I(69,	ADC,	IMMEDIATE,	1,	0,	P3(0x22,0x22,0x22)),	// ADC #IMM
	I(6A,	RORA,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// ROR ACC
	U(6B),
	I(6C,	JMP,	INDABS,		0,	0,	P3(0x05,0x05,0x05)),	// JMP (ABS)
	I(6D,	ADC,	ABS,		1,	0,	P3(0x24,0x24,0x24)),	// ADC ABS
	I(6E,	ROR,	ABS,		1,	1,	P3(0x06,0x06,0x06)),	// ROR ABS
	I(6F,	BBR,	PZ,		1,	0,	P1(0xC5)),		// BBR6 PZ,RELATIVE

//	  x	inst	address		load	store
	I(70,	BVS,	RELATIVE,	0,	0,	P3(0xC2,0xC2,0xC2)),	// BVS RELATIVE
	I(71,	ADC,	INDPZY,		1,	0,	P3(0xA5,0xA5,0xA5)),	// ADC (PZ),Y
	I(72,	ADC,	INDPZ,		1,	0,	P2(0x25,0x25)),		// ADC (PZ)
	U(73),
	I(74,	STZ,	PZX,		0,	1,	P2(0x04,0x04)),		// STZ PZ,X
	I(75,	ADC,	PZX,		1,	0,	P3(0x24,0x24,0x24)),	// ADC PZ,X
	I(76,	ROR,	PZX,		1,	1,	P3(0x06,0x06,0x06)),	// ROR PZ,X
	I(77,	RMB,	PZ,		1,	1,	P1(0x05)),		// RMB7 PZ
	I(78,	SEI,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// SEI
	I(79,	ADC,	ABSY,		1,	0,	P3(0xA4,0xA4,0xA4)),	// ADC ABS,Y
	I(7A,	PLY,	NONE,		0,	0,	P2(0x04,0x04)),		// PLY
	U(7B),
	I(7C,	JMP,	INDABSX,	0,	0,	P2(0x06,0x06)),		// JMP (ABS,X)
	I(7D,	ADC,	ABSX,		1,	0,	P3(0xA4,0xA4,0xA4)),	// ADC ABS,X
	I(7E,	ROR,	ABSX,		1,	1,	P3(0x07,0x07,0x07)),	// ROR ABS,X
	I(7F,	BBR,	PZ,		1,	0,	P1(0xC5)),		// BBR7 PZ,RELATIVE

//	  x	inst	address		load	store
	I(80,	BRA,	RELATIVE,	0,	0,	P2(0x83,0x83)),		// BRA RELATIVE
	I(81,	STA,	INDPZX,		0,	1,	P3(0x06,0x06,0x06)),	// STA (PZ,X)
	U(82),
	U(83),
	I(84,	STY,	PZ,		0,	1,	P3(0x03,0x03,0x03)),	// STY PZ
	I(85,	STA,	PZ,		0,	1,	P3(0x03,0x03,0x03)),	// STA PZ
	I(86,	STX,	PZ,		0,	1,	P3(0x03,0x03,0x03)),	// STX PZ
	I(87,	SMB,	PZ,		1,	1,	P1(0x05)),		// SMB0 PZ
	I(88,	DEY,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// DEY
	I(89,	BIT,	IMMEDIATE,	1,	0,	P2(0x03,0x03)),		// BIT #IMM
	I(8A,	TXA,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// TXA
	U(8B),
	I(8C,	STY,	ABS,		0,	1,	P3(0x04,0x04,0x04)),	// STY ABS
	I(8D,	STA,	ABS,		0,	1,	P3(0x04,0x04,0x04)),	// STA ABS
	I(8E,	STX,	ABS,		0,	1,	P3(0x04,0x04,0x04)),	// STX ABS
	I(8F,	BBS,	PZ,		1,	0,	P1(0xC5)),		// BBS0 PZ,RELATIVE

//	  x	inst	address		load	store
	I(90,	BCC,	RELATIVE,	0,	0,	P3(0xC2,0xC2,0xC2)),	// BCC RELATIVE
	I(91,	STA,	INDPZY,		0,	1,	P3(0x06,0x06,0x06)),	// STA (PZ),Y
	I(92,	STA,	INDPZ,		0,	1,	P2(0x05,0x05)),		// STA (PZ)
	U(93),
	I(94,	STY,	PZX,		0,	1,	P3(0x04,0x04,0x04)),	// STY PZ,X
	I(95,	STA,	PZX,		0,	1,	P3(0x04,0x04,0x04)),	// STA PZ,X
	I(96,	STX,	PZY,		0,	1,	P3(0x04,0x04,0x04)),	// STX PZ,Y
	I(97,	SMB,	PZ,		1,	1,	P1(0x05)),		// SMB1 PZ
	I(98,	TYA,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// TYA
	I(99,	STA,	ABSY,		0,	1,	P3(0x05,0x05,0x05)),	// STA ABS,Y
	I(9A,	TXS,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// TXS
	U(9B),
	I(9C,	STZ,	ABS,		0,	1,	P2(0x04,0x04)),		// STZ ABS
	I(9D,	STA,	ABSX,		0,	1,	P3(0x05,0x05,0x05)),	// STA ABS,X
	I(9E,	STZ,	ABSX,		0,	1,	P2(0x05,0x05)),		// STZ ABS,X
	I(9F,	BBS,	PZ,		1,	0,	P1(0xC5)),		// BBS1 PZ,RELATIVE

//	  x	inst	address		load	store
	I(A0,	LDY,	IMMEDIATE,	1,	0,	P3(0x02,0x02,0x02)),	// LDY #IMM
	I(A1,	LDA,	INDPZX,		1,	0,	P3(0x06,0x06,0x06)),	// LDA (PZ,X)
	I(A2,	LDX,	IMMEDIATE,	1,	0,	P3(0x02,0x02,0x02)),	// LDX #IMM
	U(A3),
	I(A4,	LDY,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// LDY PZ
	I(A5,	LDA,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// LDA PZ
	I(A6,	LDX,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// LDX PZ
	I(A7,	SMB,	PZ,		1,	1,	P1(0x05)),		// SMB2 PZ
	I(A8,	TAY,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// TAY
	I(A9,	LDA,	IMMEDIATE,	1,	0,	P3(0x02,0x02,0x02)),	// LDA #IMM
	I(AA,	TAX,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// TAX
	U(AB),
	I(AC,	LDY,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// LDY ABS
	I(AD,	LDA,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// LDA ABS
	I(AE,	LDX,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// LDX ABS
	I(AF,	BBS,	PZ,		1,	0,	P1(0xC5)),		// BBS2 PZ,RELATIVE

//	  x	inst	address		load	store
	I(B0,	BCS,	RELATIVE,	0,	0,	P3(0xC2,0xC2,0xC2)),	// BCS RELATIVE
	I(B1,	LDA,	INDPZY,		1,	0,	P3(0x85,0x85,0x85)),	// LDA (PZ),Y
	I(B2,	LDA,	INDPZ,		1,	0,	P2(0x05,0x05)),		// LDA (PZ)
	U(B3),
	I(B4,	LDY,	PZX,		1,	0,	P3(0x04,0x04,0x04)),	// LDY PZ,X
	I(B5,	LDA,	PZX,		1,	0,	P3(0x04,0x04,0x04)),	// LDA PZ,X
	I(B6,	LDX,	PZY,		1,	0,	P3(0x04,0x04,0x04)),	// LDX PZ,Y
	I(B7,	SMB,	PZ,		1,	1,	P1(0x05)),		// SMB3 PZ
	I(B8,	CLV,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// CLV
	I(B9,	LDA,	ABSY,		1,	0,	P3(0x84,0x84,0x84)),	// LDA ABS,Y
	I(BA,	TSX,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// TSX
	U(BB),
	I(BC,	LDY,	ABSX,		1,	0,	P3(0x84,0x84,0x84)),	// LDY ABS,X
	I(BD,	LDA,	ABSX,		1,	0,	P3(0x84,0x84,0x84)),	// LDA ABS,X
	I(BE,	LDX,	ABSY,		1,	0,	P3(0x84,0x84,0x84)),	// LDX ABS,Y
	I(BF,	BBS,	PZ,		1,	0,	P1(0xC5)),		// BBS3 PZ,RELATIVE

//	  x	inst	address		load	store
	I(C0,	CPY,	IMMEDIATE,	1,	0,	P3(0x02,0x02,0x02)),	// CPY #IMM
	I(C1,	CMP,	INDPZX,		1,	0,	P3(0x06,0x06,0x06)),	// CMP (PZ,X)
	U(C2),
	U(C3),
	I(C4,	CPY,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// CPY PZ
	I(C5,	CMP,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// CMP PZ
	I(C6,	DEC,	PZ,		1,	1,	P3(0x05,0x05,0x05)),	// DEC PZ
	I(C7,	SMB,	PZ,		1,	1,	P1(0x05)),		// SMB4 PZ
	I(C8,	INY,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// INY
	I(C9,	CMP,	IMMEDIATE,	1,	0,	P3(0x02,0x02,0x02)),	// CMP #IMM
	I(CA,	DEX,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// DEX
	I(CB,	WAI,	NONE,		0,	0,	P1(0x03)),		// WAI
	I(CC,	CPY,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// CPY ABS
	I(CD,	CMP,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// CMP ABS
	I(CE,	DEC,	ABS,		1,	1,	P3(0x06,0x06,0x06)),	// DEC ABS
	I(CF,	BBS,	PZ,		1,	0,	P1(0xC5)),		// BBS4 PZ,RELATIVE

//	  x	inst	address		load	store
	I(D0,	BNE,	RELATIVE,	0,	0,	P3(0xC2,0xC2,0xC2)),	// BNE RELATIVE
	I(D1,	CMP,	INDPZY,		1,	0,	P3(0x85,0x85,0x85)),	// CMP (PZ),Y
	I(D2,	CMP,	INDPZ,		1,	0,	P2(0x05,0x05)),		// CMP (PZ)
	U(D3),
	U(D4),
	I(D5,	CMP,	PZX,		1,	0,	P3(0x04,0x04,0x04)),	// CMP PZ,X
	I(D6,	DEC,	PZX,		1,	1,	P3(0x06,0x06,0x06)),	// DEC PZ,X
	I(D7,	SMB,	PZ,		1,	1,	P1(0x05)),		// SMB5 PZ
	I(D8,	CLD,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// CLD
	I(D9,	CMP,	ABSY,		1,	0,	P3(0x84,0x84,0x84)),	// CMP ABS,Y
	I(DA,	PHX,	NONE,		0,	0,	P2(0x03,0x03)),		// PHX
	I(DB,	STP,	NONE,		0,	0,	P1(0x03)),		// STP
	U(DC),
	I(DD,	CMP,	ABSX,		1,	0,	P3(0x84,0x84,0x84)),	// CMP ABS,X
	I(DE,	DEC,	ABSX,		1,	1,	P3(0x07,0x07,0x07)),	// DEC ABS,X
	I(DF,	BBS,	PZ,		1,	0,	P1(0xC5)),		// BBS5 PZ,RELATIVE

//	  x	inst	address		load	store
	I(E0,	CPX,	IMMEDIATE,	1,	0,	P3(0x02,0x02,0x02)),	// CPX #IMM
	I(E1,	SBC,	INDPZX,		1,	0,	P3(0x26,0x26,0x26)),	// SBC (PZ,X)
	U(E2),
	U(E3),
	I(E4,	CPX,	PZ,		1,	0,	P3(0x03,0x03,0x03)),	// CPX PZ
	I(E5,	SBC,	PZ,		1,	0,	P3(0x23,0x23,0x23)),	// SBC PZ
	I(E6,	INC,	PZ,		1,	1,	P3(0x05,0x05,0x05)),	// INC PZ
	I(E7,	SMB,	PZ,		1,	1,	P1(0x05)),		// SMB6 PZ
	I(E8,	INX,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// INX
	I(E9,	SBC,	IMMEDIATE,	1,	0,	P3(0x22,0x22,0x22)),	// SBC #IMM
	I(EA,	NOP,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// NOP
	U(EB),
	I(EC,	CPX,	ABS,		1,	0,	P3(0x04,0x04,0x04)),	// CPX ABS
	I(ED,	SBC,	ABS,		1,	0,	P3(0x24,0x24,0x24)),	// SBC ABS
	I(EE,	INC,	ABS,		1,	1,	P3(0x06,0x06,0x06)),	// INC ABS
	I(EF,	BBS,	PZ,		1,	0,	P1(0xC5)),		// BBS6 PZ,RELATIVE

//	  x	inst	address		load	store
	I(F0,	BEQ,	RELATIVE,	0,	0,	P3(0xC2,0xC2,0xC2)),	// BEQ RELATIVE
	I(F1,	SBC,	INDPZY,		1,	0,	P3(0xA5,0xA5,0xA5)),	// SBC (PZ),Y
	I(F2,	SBC,	INDPZ,		1,	0,	P2(0x25,0x25)),		// SBC (PZ)
	U(F3),
	U(F4),
	I(F5,	SBC,	PZX,		1,	0,	P3(0x24,0x24,0x24)),	// SBC PZ,X
	I(F6,	INC,	PZX,		1,	1,	P3(0x06,0x06,0x06)),	// INC PZ,X
	I(F7,	SMB,	PZ,		1,	1,	P1(0x05)),		// SMB7 PZ
	I(F8,	SED,	NONE,		0,	0,	P3(0x02,0x02,0x02)),	// SED
	I(F9,	SBC,	ABSY,		1,	0,	P3(0xA4,0xA4,0xA4)),	// SBC ABS,y
	I(FA,	PLX,	NONE,		0,	0,	P2(0x04,0x04)),		// PLX
	U(FB),
	U(FC),
	I(FD,	SBC,	ABSX,		1,	0,	P3(0xA4,0xA4,0xA4)),	// SBC ABS,X
	I(FE,	INC,	ABSX,		1,	1,	P3(0x07,0x07,0x07)),	// INC ABS,X
	I(FF,	BBS,	PZ,		1,	0,	P1(0xC5)),		// BBS7 PZ,RELATIVE
};
#undef	I
#undef	U


static void *__thread_startup(void *data)
{
	emu65C02 *emu = reinterpret_cast<emu65C02 *>(data);
	emu->Run();
	return 0;
}


emu65C02::emu65C02(processor_t processor, Callback& callback)
	: f_callback(callback)
{
	int			i, r;
	pthread_condattr_t	thread_condattr;
	pthread_mutexattr_t	thread_mutexattr;
	pthread_attr_t		thread_attr;

	// at the start we have 1 reference
	f_refcount = 1;

	for(i = 0; i < 256; ++i) {
		if(g_inst[i].f_code != i) {
			fprintf(stderr, "ERROR: invalid 6502 instruction set table at 0x%02X.\n", i);
			exit(1);
		}
	}

	f_fetchaddr[ADDR_NONE     ] = &emu65C02::FetchAddress_NONE;
	f_fetchaddr[ADDR_IMMEDIATE] = &emu65C02::FetchAddress_IMMEDIATE;
	f_fetchaddr[ADDR_PZ       ] = &emu65C02::FetchAddress_PZ;
	f_fetchaddr[ADDR_PZX      ] = &emu65C02::FetchAddress_PZX;
	f_fetchaddr[ADDR_PZY      ] = &emu65C02::FetchAddress_PZY;
	f_fetchaddr[ADDR_INDPZ    ] = &emu65C02::FetchAddress_INDPZ;
	f_fetchaddr[ADDR_INDPZX   ] = &emu65C02::FetchAddress_INDPZX;
	f_fetchaddr[ADDR_INDPZY   ] = &emu65C02::FetchAddress_INDPZY;
	f_fetchaddr[ADDR_ABS      ] = &emu65C02::FetchAddress_ABS;
	f_fetchaddr[ADDR_ABSX     ] = &emu65C02::FetchAddress_ABSX;
	f_fetchaddr[ADDR_ABSY     ] = &emu65C02::FetchAddress_ABSY;
	//f_fetchaddr[ADDR_INDABS ] = processor specific
	f_fetchaddr[ADDR_INDABSX  ] = &emu65C02::FetchAddress_INDABSX;
	f_fetchaddr[ADDR_RELATIVE ] = &emu65C02::FetchAddress_RELATIVE;

	f_processor = processor;

	if(f_processor == CPU_6502) {
		f_fetchaddr[ADDR_INDABS] = &emu65C02::FetchAddress_INDABS6502;
	}
	else {
		f_fetchaddr[ADDR_INDABS] = &emu65C02::FetchAddress_INDABS;
	}

	// Note that a Reset() is mandatory and the registers will be
	// re-initialized by that process
	f_regs.f_pc = 0;
	f_regs.f_a = 0;
	f_regs.f_x = 0;
	f_regs.f_y = 0;
	f_regs.f_s = 0xFF;
	f_regs.f_p = 0x34;

	f_current_instruction = 0xDB;	// there isn't a RESET instruction... use the STP instead
	f_temp = 0;
	f_addr = 0;
	f_stop = true;		// stopped until RESET is pressed
	f_wait = false;
	f_trace = false;
	//f_stats -- auto-init
	f_mhz = 1;
	f_speed = SPEED_PROCESSOR;
	f_cycles = 0;
	f_waited_cycles = 0;

	//f_callback -- initialized in the declaration

	// the emulator thread & messages
	f_quit = false;
	f_nmi = false;
	f_reset = false;
	f_interrupt = false;

	r = pthread_condattr_init(&thread_condattr);
	if(r != 0) {
		// the thread couldn't be created
		throw errmsg("Can't initialize pthread condition attributes");
	}
	r = pthread_cond_init(&f_cond, &thread_condattr);
	pthread_condattr_destroy(&thread_condattr);
	if(r != 0) {
		// the thread couldn't be created
		throw errmsg("Can't create thread condition");
	}

	r = pthread_mutexattr_init(&thread_mutexattr);
	if(r != 0) {
		// the thread couldn't be created
		pthread_cond_destroy(&f_cond);
		throw errmsg("Can't initialize pthread mutex attributes");
	}
	r = pthread_mutex_init(&f_mutex, &thread_mutexattr);
	pthread_mutexattr_destroy(&thread_mutexattr);
	if(r != 0) {
		// the thread couldn't be created
		pthread_cond_destroy(&f_cond);
		throw errmsg("Can't create thread mutex");
	}

	r = pthread_attr_init(&thread_attr);
	if(r != 0) {
		// the thread couldn't be created
		pthread_cond_destroy(&f_cond);
		pthread_mutex_destroy(&f_mutex);
		throw errmsg("Can't initialize pthread attributes");
	}
	r = pthread_create(&f_thread, &thread_attr, __thread_startup, this);
	pthread_attr_destroy(&thread_attr);
	if(r != 0) {
		// the thread couldn't be created
		pthread_cond_destroy(&f_cond);
		pthread_mutex_destroy(&f_mutex);
		throw errmsg("Can't create thread");
	}
}


emu65C02::~emu65C02()
{
	// wait for the thread to quit
	pthread_mutex_lock(&f_mutex);
	f_quit = true;
	pthread_cond_signal(&f_cond);
	pthread_mutex_unlock(&f_mutex);
	pthread_join(f_thread, 0);

	pthread_cond_destroy(&f_cond);
	pthread_mutex_destroy(&f_mutex);
}


unsigned int emu65C02::AddRef(void) const
{
	return ++f_refcount;
}


unsigned int emu65C02::Release(void) const
{
	unsigned int r = --f_refcount;

	if(!r) {
		delete this;
	}

	return r;
}





void emu65C02::Run(void)
{
	unsigned short	op_pc, op_end;
	unsigned char	i, cycles;
	const inst_t	*inst;
	trace_t		trace;
	bool		waited, page_cross;
	struct timeval	current_time;
	struct timezone	unused_tz;
	long long	tv_start, tv_current, diff;
	unsigned int	counter;

	gettimeofday(&f_start_time, &unused_tz);
	counter = 0;

	for(;;) {
		// first or after a STOP we wait for a RESET
		// or after a WAIT we wait for an interrupt
		waited = f_stop || f_wait;
		while(f_stop || f_wait) {
			if(f_quit) {
				return;
			}
			pthread_mutex_lock(&f_mutex);
			pthread_cond_wait(&f_cond, &f_mutex);
			pthread_mutex_unlock(&f_mutex);
		}

		if(f_quit) {
			return;
		}

#if 0
		// if not fast CPU we want to wait for a few cycles
		if(!waited) {
			pthread_mutex_lock(&f_mutex);
			pthread_cond_timedwait(&f_cond, &f_mutex);
			pthread_mutex_unlock(&f_mutex);
		}
#endif

		if(f_reset) {
			// NOTE: In most 6502, the registers are randomly initialized.
			//	 So what's here is not conventional! only some of the
			//	 bits in the status register should be set to a
			//	 specific value (XX1101XX). I'm not too sure, it could
			//	 be that the stack pointer was also NOT initialized on
			//	 a reset...
			f_callback.OnReset();

			f_regs.f_pc = LoadWord(0xFFFC);
			f_regs.f_a = f_processor;
			f_regs.f_x = 0;
			f_regs.f_y = 0;
			f_regs.f_s = 0xFF;
			f_regs.f_p = 0x34;

			f_nmi = false;
			f_interrupt = false;
			f_reset = false;
		}
		else if(f_nmi) {
			f_callback.OnNMI();

			PushWord(f_regs.f_pc);
			Push(f_regs.f_p.AsChar());
			f_regs.f_pc = LoadWord(0xFFFA);
			f_regs.f_p.f_brk = 0;

			f_nmi = false;
			f_interrupt = false;
		}
		else if(f_interrupt) {
			f_callback.OnInterrupt();

			PushWord(f_regs.f_pc);
			Push(f_regs.f_p.AsChar());
			f_regs.f_pc = LoadWord(0xFFFE);
			f_regs.f_p.f_brk = 0;

			f_interrupt = false;
		}

		trace.f_regs = f_regs;

		// read the next instruction
		f_current_instruction = Load(f_regs.f_pc);
		++f_regs.f_pc;
		inst = g_inst + f_current_instruction;

		if((f_stats.f_inst_used[f_current_instruction / 8] & (1 << (f_current_instruction & 7))) == 0) {
			++f_stats.f_inst_used_count;
			f_stats.f_inst_used[f_current_instruction / 8] |= 1 << (f_current_instruction & 7);
		}

		// valid on this processor?
		cycles = inst->f_cycles[f_processor];
		if(cycles == 0) {
			// NOTE: should we compute some lost cycles here?
			f_stats.f_unknown_opcodes++;

			trace.f_end = f_regs.f_pc;
			Trace(inst, trace);
			// skip undefined operations
			// TODO: we probably want some -debug flag...
			fflush(stdout);
fprintf(stderr, "Found undefined instruction 0x%02X at 0x%04X\n", f_current_instruction, f_regs.f_pc - 1);

			f_callback.OnUnknownOpcode(f_regs.f_pc);
			continue;
		}

		// adc/sbc in decimal?
		if((cycles & 0x20) != 0 && f_regs.f_p.f_decimal) {
			cycles++;
		}

		f_stats.f_instructions++;

		// get address (saved in emu65C02::f_addr)
		page_cross = (this->*f_fetchaddr[inst->f_addr])();
		trace.f_addr = f_addr;

		// if necessary, load data
		if(inst->f_load) {
			f_temp = Load(f_addr);
		}

		trace.f_end = f_regs.f_pc;

		// it seems to me that we lose a cycle whenever PCH is incremented
		// while reading an instruction
		if((f_regs.f_pc >> 8) != (trace.f_regs.f_pc >> 8)) {
			cycles++;
		}

		// execute instruction
		(this->*inst->f_inst)();

		// if necessary, save data
		if(inst->f_store) {
			Store(f_addr, f_temp);
		}

		switch(cycles & 0xC0) {
		case 0x80:		// extra cycle in case of page cross
			if(page_cross) {
				cycles++;
			}
			break;

		case 0xC0:		// relative extra-cycles are added only if the branch occured
			if(trace.f_end != f_regs.f_pc) {
				cycles++;	// branch taken +1
				if(page_cross) {
					cycles++;	// branch taken & page cross
				}
			}
			break;

		}

		// increment the cycles now that the instruction is over
		trace.f_cycles = (unsigned int) cycles;
		cycles &= 0x0F;
		f_cycles += cycles;
		f_stats.f_total_cycles += cycles;
		f_stats.f_cycles += cycles;

		Trace(inst, trace);

		f_callback.OnInstruction();

		// now synchronize (TODO: the current timer will break around 2068)
		gettimeofday(&current_time, &unused_tz);
		tv_start   = ((long long) f_start_time.tv_sec * 1000000) + (long long) f_start_time.tv_usec;
		tv_current = ((long long) current_time.tv_sec * 1000000) + (long long) current_time.tv_usec;
		diff = f_cycles / f_mhz - (tv_current - tv_start);

		// in case your computer gettimeofday() doesn't work I have a counter...
		if(diff > 0 || counter >= 100000 * f_mhz) {
//printf("Cycles: %lld, %lld (%lld - %lld) -> %lld\n", f_cycles, tv_current - tv_start, tv_current, tv_start, diff);
//printf("sec + usec: %ld, %ld\n", current_time.tv_sec, current_time.tv_usec);
			if(diff <= 0) {
				diff = 1;
//printf("Ooooooooooooops!\n");		// in case the counter took over...
			}
			usleep(diff);
			counter = 0;
		}
		else {
			++counter;
		}
	}
}


void emu65C02::SetTrace(bool trace)
{
	f_trace = trace;
}


void emu65C02::Trace(const inst_t *inst, trace_t& trace)
{
	static const char	cycles[16 + 1] = "S?D?????I?A?R???";
	int		p, len;
	unsigned char	op1, op2;
	char		buf[256];

	if(!f_trace) {
		return;
	}

	p = snprintf(buf, sizeof(buf), "%04X: %02X ", trace.f_regs.f_pc, f_current_instruction);
	if(trace.f_end - trace.f_regs.f_pc > 1) {
		op1 = Load(trace.f_regs.f_pc + 1);
		p += snprintf(buf + p, sizeof(buf) - p, "%02X ", op1);
		if(trace.f_end - trace.f_regs.f_pc > 2) {
			op2 = Load(trace.f_regs.f_pc + 2);
			p += snprintf(buf + p, sizeof(buf) - p, "%02X ", op2);
		}
		else {
			p += snprintf(buf + p, sizeof(buf) - p, "   ");
		}
	}
	else {
		p += snprintf(buf + p, sizeof(buf) - p, "      ");
	}
	p += snprintf(buf + p, sizeof(buf) - p, " %.3s", inst->f_name);

	switch(inst->f_addr) {
	default: // ADDR_NONE (also includes ADDR_MAX)
		break;

	case ADDR_IMMEDIATE:
		p += snprintf(buf + p, sizeof(buf) - p, " #$%02X", op1);
		break;

	case ADDR_PZ:
		p += snprintf(buf + p, sizeof(buf) - p, " $%02X", op1);
		break;

	case ADDR_PZX:
		p += snprintf(buf + p, sizeof(buf) - p, " $%02X,X", op1);
		break;

	case ADDR_PZY:
		p += snprintf(buf + p, sizeof(buf) - p, " $%02X,Y", op1);
		break;

	case ADDR_INDPZ:
		p += snprintf(buf + p, sizeof(buf) - p, " ($%02X)", op1);
		break;

	case ADDR_INDPZX:
		p += snprintf(buf + p, sizeof(buf) - p, " ($%02X,X)", op1);
		break;

	case ADDR_INDPZY:
		p += snprintf(buf + p, sizeof(buf) - p, " ($%02X),Y", op1);
		break;

	case ADDR_ABS:
		p += snprintf(buf + p, sizeof(buf) - p, " $%04X", op1 + (op2 << 8));
		break;

	case ADDR_ABSX:
		p += snprintf(buf + p, sizeof(buf) - p, " $%04X,X", op1 + (op2 << 8));
		break;

	case ADDR_ABSY:
		p += snprintf(buf + p, sizeof(buf) - p, " $%04X,Y", op1 + (op2 << 8));
		break;

	case ADDR_INDABS:
		p += snprintf(buf + p, sizeof(buf) - p, " ($%04X)", op1 + (op2 << 8));
		break;

	case ADDR_INDABSX:
		p += snprintf(buf + p, sizeof(buf) - p, " ($%04X,X)", op1 + (op2 << 8));
		break;

	case ADDR_RELATIVE:
		p += snprintf(buf + p, sizeof(buf) - p, " $%04X", (trace.f_end + (signed char) op1) & 0x0FFFF);
		break;

	}

	while(p < 28) {
		buf[p] = ' ';
		p++;
	}

	p += snprintf(buf + p, sizeof(buf) - p,
		"A:$%02X  X:$%02X  Y:$%02X  S:$%02X  M:$",
		trace.f_regs.f_a, trace.f_regs.f_x,
		trace.f_regs.f_y, trace.f_regs.f_s);
	if(inst->f_load || inst->f_store) {
		p += snprintf(buf + p, sizeof(buf) - p, "%04X", trace.f_addr);
	}
	else {
		p += snprintf(buf + p, sizeof(buf) - p, "----");
	}
	p += snprintf(buf + p, sizeof(buf) - p,
		"  P:%c%c1%c%c%c%c%c",
		trace.f_regs.f_p.f_negative  ? 'N' : '-',
		trace.f_regs.f_p.f_overflow  ? 'V' : '-',
		trace.f_regs.f_p.f_brk       ? 'B' : '-',
		trace.f_regs.f_p.f_decimal   ? 'D' : '-',
		trace.f_regs.f_p.f_interrupt ? 'I' : '-',
		trace.f_regs.f_p.f_zero      ? 'Z' : '-',
		trace.f_regs.f_p.f_carry     ? 'C' : '-');

	p += snprintf(buf + p, sizeof(buf) - p,
		"  C:%c/%d (%d)",
		cycles[trace.f_cycles >> 4],
		trace.f_cycles & 0x0F,
		f_stats.f_cycles);

	if(p >= sizeof(buf)) {
		p--;
	}
	buf[p] = '\0';

	f_callback.OnTrace(buf);
}


void emu65C02::NMI(void)
{
	pthread_mutex_lock(&f_mutex);
	f_nmi = true;
	f_wait = false;
	pthread_cond_signal(&f_cond);
	pthread_mutex_unlock(&f_mutex);
}


void emu65C02::Reset(void)
{
	pthread_mutex_lock(&f_mutex);
	f_reset = true;
	f_stop = false;
	f_wait = false;
	pthread_cond_signal(&f_cond);
	pthread_mutex_unlock(&f_mutex);
}


void emu65C02::Interrupt(void)
{
	pthread_mutex_lock(&f_mutex);
	f_interrupt = true;
	f_wait = false;
	pthread_cond_signal(&f_cond);
	pthread_mutex_unlock(&f_mutex);
}


void emu65C02::GetStats(cpu_stats_t& stats, bool reset) const
{
	stats = f_stats;

	if(reset) {
		f_stats.Reset();
	}
}



unsigned int emu65C02::GetMhz(void) const
{
	return f_mhz;
}


void emu65C02::SetMhz(unsigned int mhz)
{
	struct timezone	unused_tz;

	// A speed of about 10 is the max. on a Celeron 2500Mhz
	if(mhz > 0 && mhz < 100) {
		// reset everything in link with speed
		f_mhz = mhz;
		gettimeofday(&f_start_time, &unused_tz);
		f_cycles = 0;
	}
}


emu65xx::speed_t emu65C02::GetSpeed(void) const
{
	return f_speed;
}


void emu65C02::SetSpeed(speed_t speed)
{
	f_speed = speed;
}



};			// namespace emulator_8bits
// vim: ts=8
