/* emu65xx.h -- written by Alexis WILKE (c) 2005-2006 */

/**********************************************************************
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**********************************************************************/


#include	"emu.h"

#include	<stdio.h>
//#include	<stdlib.h>
//#include	<string.h>

namespace emulator_8bits
{



void emu65C02::inst_BBR(void)	// BBR# PZ,RELATIVE
{
	signed char rel = Load(f_regs.f_pc);
	f_regs.f_pc++;
	if(((1 << (f_current_instruction >> 4)) & f_temp) == 0) {
		f_regs.f_pc += rel;
	}
}


void emu65C02::inst_BBS(void)	// BBS# PZ,RELATIVE
{
	signed char rel = Load(f_regs.f_pc);
	f_regs.f_pc++;
	if(((1 << ((f_current_instruction >> 4) & 7)) & f_temp) != 0) {
		f_regs.f_pc += rel;
	}
}


void emu65C02::inst_BCC(void)	// BCC RELATIVE
{
	if(f_regs.f_p.f_carry == 0) {
		f_regs.f_pc = f_addr;
	}
}


void emu65C02::inst_BCS(void)	// BCS RELATIVE
{
	if(f_regs.f_p.f_carry == 1) {
		f_regs.f_pc = f_addr;
	}
}


void emu65C02::inst_BEQ(void)	// BEQ RELATIVE
{
	if(f_regs.f_p.f_zero == 1) {
		f_regs.f_pc = f_addr;
	}
}


void emu65C02::inst_BIT(void)	// BIT
{
	f_regs.f_p.f_zero = (f_regs.f_a & f_temp) == 0;
	f_regs.f_p.f_overflow = f_temp >> 6;
	f_regs.f_p.f_negative = f_temp >> 7;
}


void emu65C02::inst_BMI(void)	// BMI RELATIVE
{
	if(f_regs.f_p.f_negative == 1) {
		f_regs.f_pc = f_addr;
	}
}


void emu65C02::inst_BNE(void)	// BNE RELATIVE
{
	if(f_regs.f_p.f_zero == 0) {
		f_regs.f_pc = f_addr;
	}
}


void emu65C02::inst_BPL(void)	// BPL RELATIVE
{
	if(f_regs.f_p.f_negative == 0) {
		f_regs.f_pc = f_addr;
	}
}


void emu65C02::inst_BRA(void)	// BRA RELATIVE
{
	f_regs.f_pc = f_addr;
}


void emu65C02::inst_BRK(void)	// BRK
{
	PushWord(f_regs.f_pc);
	Push(f_regs.f_p.AsChar());
	f_regs.f_pc = LoadWord(0xFFFE);
	f_regs.f_p.f_brk = 1;
	if(f_processor != CPU_6502) {
		f_regs.f_p.f_decimal = 0;
		// TODO: I think the interrupt flag is always set...
		//	 some documents contradict each others
		f_regs.f_p.f_interrupt = 1;
	}
}


void emu65C02::inst_BVC(void)	// BVC RELATIVE
{
	if(f_regs.f_p.f_overflow == 0) {
		f_regs.f_pc = f_addr;
	}
}


void emu65C02::inst_BVS(void)	// BVS RELATIVE
{
	if(f_regs.f_p.f_overflow == 1) {
		f_regs.f_pc = f_addr;
	}
}


void emu65C02::inst_ADC(void)	// ADC
{
	unsigned short	total;

	if(f_regs.f_p.f_decimal) {
		total = (f_regs.f_a & 0x0F) + (f_regs.f_a >> 4) * 10
			+ f_regs.f_p.f_carry
			+ (f_temp & 0x0F) + (f_temp >> 4) * 10;

		if(total >= 100) {
			f_regs.f_p.f_carry = 1;
		}
		else {
			f_regs.f_p.f_carry = 0;
		}

		total = (total % 10) + (total / 10) * 16;
	}
	else {
		total = f_regs.f_a + f_regs.f_p.f_carry + f_temp;

		if(total >= 0x100) {
			f_regs.f_p.f_carry = 1;
		}
		else {
			f_regs.f_p.f_carry = 0;
		}

		// V is set when the ADC changes the sign of the result "improperly"
		// This happens when f_regs.f_a and f_temp have the same sign and total doesn't
		if((signed char) (f_regs.f_a ^ f_temp) >= 0 && (signed char) (f_regs.f_a ^ total) < 0) {
			f_regs.f_p.f_overflow = 1;
		}
		else {
			f_regs.f_p.f_overflow = 0;
		}
	}

	f_regs.f_a = (unsigned char) total;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_AND(void)	// AND
{
	f_regs.f_a &= f_temp;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_ASL(void)	// ASL
{
	f_regs.f_p.f_carry = f_temp >> 7;
	f_temp <<= 1;
	SetLoadFlags(f_temp);
}


void emu65C02::inst_ASLA(void)	// ASL ACC
{
	f_regs.f_p.f_carry = f_regs.f_a >> 7;
	f_regs.f_a <<= 1;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_CLC(void)	// CLC
{
	f_regs.f_p.f_carry = 0;
}


void emu65C02::inst_CLD(void)	// CLD
{
	f_regs.f_p.f_decimal = 0;
}


void emu65C02::inst_CLI(void)	// CLI
{
	f_regs.f_p.f_interrupt = 0;
}


void emu65C02::inst_CLV(void)	// CLV
{
	f_regs.f_p.f_overflow = 0;
}


void emu65C02::inst_CMP(void)	// CMP
{
	unsigned short	total;

	total = f_regs.f_a - f_temp;

	if(total >= 0x100) {
		f_regs.f_p.f_carry = 0;
	}
	else {
		f_regs.f_p.f_carry = 1;
	}

	SetLoadFlags(total);
}


void emu65C02::inst_CPX(void)	// CPX
{
	unsigned short	total;

	total = f_regs.f_x - f_temp;

	if(total >= 0x100) {
		f_regs.f_p.f_carry = 0;
	}
	else {
		f_regs.f_p.f_carry = 1;
	}

	SetLoadFlags(total);
}


void emu65C02::inst_CPY(void)	// CPY
{
	unsigned short	total;

	total = f_regs.f_y - f_temp;

	if(total >= 0x100) {
		f_regs.f_p.f_carry = 0;
	}
	else {
		f_regs.f_p.f_carry = 1;
	}

	SetLoadFlags(total);
}


void emu65C02::inst_DEC(void)	// DEC
{
	f_temp--;
	SetLoadFlags(f_temp);
}


void emu65C02::inst_DECA(void)	// DEC ACC
{
	f_regs.f_a--;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_DEX(void)	// DEX
{
	f_regs.f_x--;
	SetLoadFlags(f_regs.f_x);
}


void emu65C02::inst_DEY(void)	// DEY
{
	f_regs.f_y--;
	SetLoadFlags(f_regs.f_y);
}


void emu65C02::inst_EOR(void)	// EOR
{
	f_regs.f_a ^= f_temp;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_INC(void)	// INC
{
	f_temp++;
	SetLoadFlags(f_temp);
}


void emu65C02::inst_INCA(void)	// INC ACC
{
	f_regs.f_a++;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_INX(void)	// INX
{
	f_regs.f_x++;
	SetLoadFlags(f_regs.f_x);
}


void emu65C02::inst_INY(void)	// INY
{
	f_regs.f_y++;
	SetLoadFlags(f_regs.f_y);
}


void emu65C02::inst_JMP(void)	// JMP ABS
{
	f_regs.f_pc = f_addr;
}


void emu65C02::inst_JMPIND(void)	// JMP (ABS[,X])
{
	f_regs.f_pc = f_temp + (Load(f_addr + 1) << 8);
}


void emu65C02::inst_JSR(void)	// JSR
{
	PushWord(f_regs.f_pc - 1);
	f_regs.f_pc = f_addr;
}


void emu65C02::inst_LDA(void)	// LDA
{
	f_regs.f_a = f_temp;
	SetLoadFlags(f_temp);
}


void emu65C02::inst_LDX(void)	// LDX
{
	f_regs.f_x = f_temp;
	SetLoadFlags(f_temp);
}


void emu65C02::inst_LDY(void)	// LDY
{
	f_regs.f_y = f_temp;
	SetLoadFlags(f_temp);
}


void emu65C02::inst_LSR(void)	// LSR
{
	f_regs.f_p.f_carry = f_temp & 1;
	f_temp >>= 1;
	SetLoadFlags(f_temp);
}


void emu65C02::inst_LSRA(void)	// LSR ACC
{
	f_regs.f_p.f_carry = f_regs.f_a & 1;
	f_regs.f_a >>= 1;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_NOP(void)	// NOP
{
}
	// do nothing (waste some cycles)


void emu65C02::inst_ORA(void)	// ORA
{
	f_regs.f_a |= f_temp;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_PHA(void)	// PHA
{
	Push(f_regs.f_a);
}


void emu65C02::inst_PHP(void)	// PHP
{
	Push(f_regs.f_p.AsChar());
}


void emu65C02::inst_PHX(void)	// PHX
{
	Push(f_regs.f_x);
}


void emu65C02::inst_PHY(void)	// PHY
{
	Push(f_regs.f_y);
}


void emu65C02::inst_PLA(void)	// PLA
{
	f_regs.f_a = Pop();
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_PLP(void)	// PLP
{
	f_regs.f_p = Pop();
	f_regs.f_p.f_one = 1;
	f_regs.f_p.f_brk = 1;	// TODO: I think this is always 1 here
}


void emu65C02::inst_PLX(void)	// PLX
{
	f_regs.f_x = Pop();
	SetLoadFlags(f_regs.f_x);
}


void emu65C02::inst_PLY(void)	// PLY
{
	f_regs.f_y = Pop();
	SetLoadFlags(f_regs.f_y);
}


void emu65C02::inst_RMB(void)	// RMB
{
	f_temp &= ~(1 << (f_current_instruction >> 4));
}


void emu65C02::inst_ROL(void)	// ROL
{
	unsigned char r = (f_temp << 1) + f_regs.f_p.f_carry;
	f_regs.f_p.f_carry = f_temp >> 7;
	f_temp = r;
	SetLoadFlags(r);
}


void emu65C02::inst_ROLA(void)	// ROL ACC
{
	unsigned char r = (f_regs.f_a << 1) + f_regs.f_p.f_carry;
	f_regs.f_p.f_carry = f_regs.f_a >> 7;
	f_regs.f_a = r;
	SetLoadFlags(r);
}


void emu65C02::inst_ROR(void)	// ROR
{
	unsigned char r = (f_temp >> 1) + (f_regs.f_p.f_carry << 7);
	f_regs.f_p.f_carry = f_temp & 1;
	f_temp = r;
	SetLoadFlags(r);
}


void emu65C02::inst_RORA(void)	// ROR ACC
{
	unsigned char r = (f_regs.f_a >> 1) + (f_regs.f_p.f_carry << 7);
	f_regs.f_p.f_carry = f_regs.f_a & 1;
	f_regs.f_a = r;
	SetLoadFlags(r);
}


void emu65C02::inst_RTI(void)	// RTI
{
	f_regs.f_p = Pop();
	f_regs.f_p.f_one = 1;
	f_regs.f_p.f_brk = 1;	// TODO: I think this is always 1 here
	f_regs.f_pc = PopWord();
}


void emu65C02::inst_RTS(void)	// RTS
{
	f_regs.f_pc = PopWord() + 1;
}


void emu65C02::inst_SBC(void)	// SBC
{
	unsigned short	total;

	if(f_regs.f_p.f_decimal) {
		total = (f_regs.f_a & 0x0F) + (f_regs.f_a >> 4) * 10
			- !f_regs.f_p.f_carry
			- (f_temp & 0x0F) - (f_temp >> 4) * 10;

		if(total >= 100) {
			f_regs.f_p.f_carry = 0;
		}
		else {
			f_regs.f_p.f_carry = 1;
		}

		total = (total % 10) + (total / 10) * 16;
	}
	else {

		// for SBC the carry is inverted
		total = f_regs.f_a - !f_regs.f_p.f_carry - f_temp;

		if(total >= 0x100) {
			f_regs.f_p.f_carry = 0;
		}
		else {
			f_regs.f_p.f_carry = 1;
		}

		// V is set when the SBC changes the sign of the result "improperly"
		// This happens when f_regs.f_a and f_temp have the same sign and total doesn't
		if((signed char) (f_regs.f_a ^ f_temp) >= 0 && (signed char) (f_regs.f_a ^ total) < 0) {
			f_regs.f_p.f_overflow = 1;
		}
		else {
			f_regs.f_p.f_overflow = 0;
		}
	}

	f_regs.f_a = (unsigned char) total;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_SEC(void)	// SEC
{
	f_regs.f_p.f_carry = 1;
}


void emu65C02::inst_SED(void)	// SED
{
	f_regs.f_p.f_decimal = 1;
}


void emu65C02::inst_SEI(void)	// SEI
{
	f_regs.f_p.f_interrupt = 1;
}


void emu65C02::inst_SMB(void)	// SMB
{
	f_temp |= 1 << (f_current_instruction >> 4);
}


void emu65C02::inst_STA(void)	// STA
{
	f_temp = f_regs.f_a;
}


void emu65C02::inst_STP(void)	// STP
{
	f_stop = true;
}


void emu65C02::inst_STX(void)	// STX
{
	f_temp = f_regs.f_x;
}


void emu65C02::inst_STY(void)	// STY
{
	f_temp = f_regs.f_y;
}


void emu65C02::inst_STZ(void)	// STZ
{
	f_temp = 0;
}


void emu65C02::inst_TAX(void)	// TAX
{
	f_regs.f_x = f_regs.f_a;
	SetLoadFlags(f_regs.f_x);
}


void emu65C02::inst_TAY(void)	// TAY
{
	f_regs.f_y = f_regs.f_a;
	SetLoadFlags(f_regs.f_y);
}


void emu65C02::inst_TRB(void)	// TRB
{
	f_regs.f_p.f_zero = (f_temp & f_regs.f_a) != 0;
	f_temp &= ~f_regs.f_a;
}


void emu65C02::inst_TSB(void)	// TSB
{
	f_regs.f_p.f_zero = (f_temp & f_regs.f_a) != 0;
	f_temp |= f_regs.f_a;
}


void emu65C02::inst_TSX(void)	// TSX
{
	f_regs.f_x = f_regs.f_s;
	SetLoadFlags(f_regs.f_x);
}


void emu65C02::inst_TXA(void)	// TXA
{
	f_regs.f_a = f_regs.f_x;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_TXS(void)	// TXS
{
	f_regs.f_s = f_regs.f_x;
}


void emu65C02::inst_TYA(void)	// TYA
{
	f_regs.f_a = f_regs.f_y;
	SetLoadFlags(f_regs.f_a);
}


void emu65C02::inst_WAI(void)	// WAI
{
	f_wait = 1;
}





};			// namespace emulator_8bits
// vim: ts=8
