/* emu65xx.h -- written by Alexis WILKE (c) 2005-2006 */

/**********************************************************************
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**********************************************************************/

#ifndef __EMULATOR_H__
#define __EMULATOR_H__


namespace emulator_8bits
{

// to throw error messages
class errmsg
{
public:
			errmsg(const char *msg)
			{
				int i;
				for(i = 0; i < sizeof(f_msg) - 1 && *msg != '\0'; ++i) {
					f_msg[i] = *msg++;
				}
				f_msg[i] = '\0';
			}

	const char *	getmsg(void) const;

private:
	char		f_msg[256];
};

class emu65xx
{
public:
	// different implementations of this emulator
	enum id_t {
		ID_65C02
	};

	enum processor_t {
		CPU_6502,
		CPU_65C02,
		CPU_65C02S,
		CPU_65CE02,
		CPU_max
	};

	enum speed_t {
		SPEED_FAST,
		SPEED_PROCESSOR,
		SPEED_SLOW
	};

	struct cpu_stats_t {
				cpu_stats_t(void)
				{
					f_total_cycles = 0;	// cannot be reset
					Reset();
				}
		void		Reset(void)
				{
					f_instructions = 0;
					f_unknown_opcodes = 0;
					f_cycles = 0;
					f_read = 0;
					f_write = 0;
					f_inst_used_count = 0;
					f_page_read_count = 0;
					f_page_written_count = 0;
					for(int i = 0; i < 256 / 8; ++i) {
						f_inst_used[i] = 0;
						f_page_read[i] = 0;
						f_page_written[i] = 0;
					}
				}

		unsigned int	f_instructions;
		unsigned int	f_unknown_opcodes;
		unsigned int	f_total_cycles;
		unsigned int	f_cycles;
		unsigned int	f_read;
		unsigned int	f_write;
		unsigned int	f_inst_used_count;
		unsigned int	f_page_read_count;
		unsigned int	f_page_written_count;
		unsigned char	f_inst_used[256 / 8];
		unsigned char	f_page_read[256 / 8];
		unsigned char	f_page_written[256 / 8];
	};

	class Callback
	{
	public:
		virtual void		OnNMI(void) {}
		virtual void		OnReset(void) {}
		virtual void		OnInterrupt(void) {}

		virtual void		OnLoad(unsigned short addr, unsigned char& value) = 0;
		virtual void		OnStore(unsigned short addr, unsigned char value) = 0;

		virtual void		OnInstruction(void) {}
		virtual void		OnUnknownOpcode(unsigned short& pc) {}
		virtual void		OnTrace(const char *msg) {}
	};

	static emu65xx *	Create(id_t id, processor_t processor, Callback& callback);

	virtual unsigned int	AddRef(void) const = 0;
	virtual unsigned int	Release(void) const = 0;

	virtual void		SetTrace(bool trace = true) = 0;
	virtual unsigned int	GetMhz(void) const = 0;
	virtual void		SetMhz(unsigned int mhz) = 0;
	virtual speed_t		GetSpeed(void) const = 0;
	virtual void		SetSpeed(speed_t speed) = 0;
	virtual void		GetStats(cpu_stats_t& st, bool reset = true) const = 0;

	virtual void		NMI(void) = 0;
	virtual void		Reset(void) = 0;
	virtual void		Interrupt(void) = 0;
};


};		// emulator_8bits
#endif		// __EMULATOR_H__
// vim: ts=8
